-- language-independent prelude facilities

oper

-- to construct records and tables
  SS  : Type = {s : Str} ;
  ss  : Str -> SS = \s -> {s = s} ;
  ss2 : (_,_ : Str) -> SS = \x,y -> ss (x ++ y) ;
  ss3 : (_,_ ,_: Str) -> SS = \x,y,z -> ss (x ++ y ++ z) ;

  cc2 : (_,_ : SS) -> SS = \x,y -> ss (x.s ++ y.s) ;

  SS1 : Type -> Type = \P -> {s : P => Str} ;
  ss1 : (A : Type) -> Str -> SS1 A = \_,s -> {s = table {_ => s}} ;

  SP1 : Type -> Type = \P -> {s : Str ; p : P} ;
  sp1 : (A : Type) -> Str -> A -> SP1 A = \_,s,a -> {s = s ; p = a} ;

  nonExist : Str = variants {} ;

  optStr : Str -> Str = \s -> variants {s ; []} ;

  constTable : (A,B : Type) -> B -> A => B = \_,_,b -> table {_ => b} ;
  constStr   : (A : Type) -> Str -> A => Str = \A,b -> constTable A Str b ;

  infixSS   : Str  -> SS -> SS -> SS = \f,x,y -> ss (x.s ++ f ++ y.s) ;
  prefixSS  : Str        -> SS -> SS = \f,x   -> ss (f ++ x.s) ;
  postfixSS : Str        -> SS -> SS = \f,x   -> ss (x.s ++ f) ;
  embedSS   : Str -> Str -> SS -> SS = \f,g,x -> ss (f ++ x.s ++ g) ;

-- discontinuous
  SD2 = {s1,s2 : Str} ;
  sd2 : (_,_ : Str) -> SD2 = \x,y -> {s1 = x ; s2 = y} ;

-- parentheses
  paren : Str -> Str = \s -> "(" ++ s ++ ")" ;
  parenss : SS -> SS = \s -> ss (paren s.s) ;

-- free order between two strings
  bothWays : Str -> Str -> Str = \x,y -> variants {x ++ y ; y ++ x} ;

-- parametric order between two strings
  preOrPost : Bool -> Str -> Str -> Str = \pr,x,y -> 
    if_then_else Str pr (x ++ y) (y ++ x) ;

-- Booleans

param Bool = True | False ;

oper
  if_then_else : (A : Type) -> Bool -> A -> A -> A = \_,c,d,e -> 
    case c of {
      True => d ;
      False => e
     } ;

  andB : (_,_ : Bool) -> Bool = \a,b -> if_then_else Bool a b False ;
  orB  : (_,_ : Bool) -> Bool = \a,b -> if_then_else Bool a True b ;
  notB : Bool         -> Bool = \a   -> if_then_else Bool a False True ;


-- zero, one, two, or more (elements in a list etc)

param
  ENumber = E0 | E1 | E2 | Emore ;

oper
  eNext : ENumber -> ENumber = \e -> case e of {
    E0 => E1 ; E1 => E2 ; _ => Emore} ;



-------------------------


--
--2 Enumerated parameter types 
--
-- These types are the ones found in school grammars.
-- Their parameter values are atomic.

param 
  Num = Sg | Pl ;
  Gend = NoHum | Hum ;
  Case = Nom | Gen ;
  Person = P1 | P2 | P3 ;
  Degree = Pos | Comp | Sup ;

-- For data abstraction, we define

oper 
  singular = Sg ;
  plural = Pl ;

--2 Word classes and hierarchical parameter types
--
-- Real parameter types (i.e. ones on which words and phrases depend) 
-- are often hierarchical. The alternative would be cross-products of
-- simple parameters, but this would usually overgenerate.
--

--3 Common nouns
--
-- Common nouns are inflected in number and case.

  CommonNoun : Type = {s : Num => Case => Str} ;


--
--3 Adjectives
--
-- The major division is between the comparison degrees, but it
-- is also good to leave room for adjectives that cannon be compared.
-- Such adjectives are simply strings.

  Adjective : Type = SS ;
  AdjDegr = SS1 Degree ;

--3 Verbs
--
-- We limit the grammar so far to verbs in infinitive-imperative or present tense.
-- The present tense is made to depend on person, which correspond to forms
-- in the singular; plural forms are uniformly equal to the 2nd person singular.

param
  VForm = InfImp | Indic Person ;

oper
  VerbP3 : Type = SS1 VForm ;
  Verb = VerbP3 ; 

--
--3 Pronouns
--
-- For pronouns, we need four case forms: "I" - "me" - "my" - "mine".

param
  CaseP = NomP | AccP | GenP | GenSP ; 

oper
  Pronoun  : Type = {s : CaseP => Str ; n : Num ; p : Person} ;

-- Coercions between pronoun cases and ordinaty cases.

  toCase  : CaseP -> Case = \c -> case c of {GenP => Gen ; _ => Nom} ;
  toCaseP : Case -> CaseP = \c -> case c of {Gen  => GenP ; _ => NomP} ; ---

--3 Proper names
--
-- Proper names only need two cases.

  ProperName : Type = SS1 Case ;

--3 Relative pronouns
--
-- Relative pronouns are inflected in gender (human/nonhuman), number, and case.

  RelPron : Type = {s : Gend => Num => CaseP => Str} ;



---------------------------------



--2 Nouns
--
-- For conciseness and abstraction, we define a worst-case macro for
-- noun inflection. It is used for defining special case that
-- only need one string as argument.

oper
  mkNoun  : (_,_,_,_ : Str) -> CommonNoun = 
    \man,men, mans, mens -> {s = table {
       Sg => table {Nom => man ; Gen => mans} ;
       Pl => table {Nom => men ; Gen => mens}
      }} ;

  nounReg : Str -> CommonNoun = \dog -> 
    mkNoun dog (dog + "s") (dog + "'s") (dog + "s'");

  nounS   : Str -> CommonNoun = \kiss -> 
    mkNoun kiss (kiss + "es") (kiss + "'s") (kiss + "es'") ;

  nounY   : Str -> CommonNoun = \fl -> 
    mkNoun (fl + "y") (fl + "ies") (fl + "y's") (fl + "ies'") ;

--3 Proper names
--
-- Regular proper names are inflected with "'s" in the genitive.

  nameReg : Str -> ProperName = \john -> 
    {s = table {Nom => john ; Gen => john + "'s"}} ;
   

--2 Pronouns
--
-- Here we define personal and relative pronouns.

  mkPronoun : (_,_,_,_ : Str) -> Num -> Person -> Pronoun = \I,me,my,mine,n,p -> 
    {s = table {NomP => I ; AccP => me ; GenP => my ; GenSP => mine} ; 
     n = n ; p = p} ;

  pronI = mkPronoun "I" "me" "my" "mine" Sg P1 ;
  pronYouSg = mkPronoun "you" "you" "your" "yours" Sg P2 ; -- verb form still OK
  pronHe = mkPronoun "he" "him" "his" "his" Sg P3 ;
  pronShe = mkPronoun "she" "her" "her" "hers" Sg P3 ;

  pronWe = mkPronoun "we" "us" "our" "ours" Pl P1 ;
  pronYouPl = mkPronoun "you" "you" "your" "yours" Pl P2 ;
  pronThey = mkPronoun "they" "them" "their" "theirs" Pl P3 ;

-- Relative pronouns in the accusative have the 'no pronoun' variant.
-- The simple pronouns do not really depend on number.

  relPron : RelPron = {s = table {
    NoHum => \\_ => table {
      NomP => variants {"that" ; "which"} ;
      AccP => variants {"that" ; "which" ; []} ;
      GenP => variants {"whose"} ;
      GenSP => variants {"which"}
      } ;
    Hum => \\_ => table {
      NomP => variants {"that" ; "who"} ;
      AccP => variants {"that" ; "who" ; "whom" ; []} ;
      GenP => variants {"whose"} ;
      GenSP => variants {"whom"}
      }
    }
  } ;


--3 Determiners
--
-- We have just a heuristic definition of the indefinite article.
-- There are lots of exceptions: consonantic "e" ("euphemism"), consonantic
-- "o" ("one-sided"), vocalic "u" ("umbrella").

  artIndef = pre {"a" ; 
                  "an" / strs {"a" ; "e" ; "i" ; "o" ; "A" ; "E" ; "I" ; "O" }} ;

  artDef = "the" ;

--2 Adjectives
--
-- For the comparison of adjectives, three forms are needed in the worst case.

  mkAdjDegr : (_,_,_ : Str) -> AdjDegr = \good,better,best -> 
    {s = table {Pos => good ; Comp => better ; Sup => best}} ;

  adjDegrReg : Str -> AdjDegr = \long ->
    mkAdjDegr long (long + "er") (long + "est") ;

  adjDegrY : Str -> AdjDegr = \lovel ->
    mkAdjDegr (lovel + "y") (lovel + "ier") (lovel + "iest") ;

-- Many adjectives are 'inflected' by adding a comparison word.

  adjDegrLong : Str -> AdjDegr = \ridiculous ->
    mkAdjDegr ridiculous ("more" ++ ridiculous) ("most" ++ ridiculous) ;

-- simple adjectives are just strings

  simpleAdj : Str -> Adjective = ss ;

--3 Verbs
--
-- Except for "be", the worst case needs two forms.

  mkVerbP3 : (_,_: Str) -> VerbP3 = \goes,go -> 
    {s = table {InfImp => go ; Indic P3 => goes ; Indic _ => go}} ;

  regVerbP3 : Str -> VerbP3 = \walk -> 
    mkVerbP3 (walk + "s") walk ;

  verbP3s   : Str -> VerbP3 = \kiss -> 
    mkVerbP3 (kiss + "es") kiss ;

  verbP3y   : Str -> VerbP3 = \fl -> 
    mkVerbP3 (fl + "ies") (fl + "y") ;

  verbP3Have = mkVerbP3 "has" "have" ;

  verbP3Do   = verbP3s "do" ;

  verbBe : VerbP3 = {s = table {
    InfImp => "be" ; 
    Indic P1 => "am" ; 
    Indic P2 => "are" ;
    Indic P3 => "is"
    }} ;

-- The optional negation contraction is a useful macro e.g. for "do".

  contractNot : Str -> Str = \is -> variants {is ++ "not" ; is + "n't"} ;

  dont = contractNot (verbP3Do.s ! InfImp) ;



package CO = {

param
  ListSize = TwoElem | ManyElem ;

oper 
  SS = {s : Str} ; ----

  ListX = {s1,s2 : Str} ;

  twoStr : (x,y : Str) -> ListX = \x,y -> 
    {s1 = x ; s2 = y} ; 
  consStr : Str -> ListX -> Str -> ListX = \comma,xs,x -> 
    {s1 = xs.s1 ++ comma ++ xs.s2 ; s2 = x } ; 

  twoSS : (_,_ : SS) -> ListX = \x,y -> 
    twoStr x.s y.s ;
  consSS : Str -> ListX -> SS -> ListX = \comma,xs,x -> 
    consStr comma xs x.s ;

  Conjunction : Type = SS ;
  ConjunctionDistr : Type = {s1 : Str ; s2 : Str} ;

  conjunctX : Conjunction -> ListX -> Str = \or,xs ->
    xs.s1 ++ or.s ++ xs.s2 ;

  conjunctDistrX : ConjunctionDistr -> ListX -> Str = \or,xs ->
    or.s1 ++ xs.s1 ++ or.s2 ++ xs.s2 ;

  -- all this lifted to tables

  ListTable : Type -> Type = \P -> {s1,s2 : P => Str} ; 

  twoTable : (P : Type) -> (_,_ : {s : P => Str}) -> ListTable P = \_,x,y ->
    {s1 = x.s ; s2 = y.s} ; 

  consTable : (P : Type) -> Str -> ListTable P -> {s : P => Str} -> ListTable P = 
    \_,c,xs,x ->
    {s1 = \\o => xs.s1 ! o ++ c ++ xs.s2 ! o ; s2 = x.s} ; 

  conjunctTable : (P : Type) -> Conjunction -> ListTable P -> {s : P => Str} = 
    \_,or,xs ->
    {s = \\p => xs.s1 ! p ++ or.s ++ xs.s2 ! p} ;

  conjunctDistrTable : 
    (P : Type) -> ConjunctionDistr -> ListTable P -> {s : P => Str} = \_,or,xs ->
    {s = \\p => or.s1++ xs.s1 ! p ++ or.s2 ++ xs.s2 ! p} ;

  -- ... and to two-argument tables: how clumsy! ---

  ListTable2 : Type -> Type -> Type = \P,Q -> 
    {s1,s2 : P => Q => Str} ; 

  twoTable2 : (P,Q : Type) -> (_,_ : {s : P => Q => Str}) -> ListTable2 P Q = 
    \_,_,x,y ->
    {s1 = x.s ; s2 = y.s} ; 

  consTable2 : 
    (P,Q : Type) -> Str -> ListTable2 P Q -> {s : P => Q => Str} -> ListTable2 P Q =
     \_,_,c,xs,x ->
    {s1 = \\p,q => xs.s1 ! p ! q ++ c ++ xs.s2 !p!q ; s2 = x.s} ; 

  conjunctTable2 : 
    (P,Q : Type) -> Conjunction -> ListTable2 P Q -> {s : P => Q => Str} = 
    \_,_,or,xs ->
    {s = \\p,q => xs.s1 ! p ! q ++ or.s ++ xs.s2 ! p ! q} ;

  conjunctDistrTable2 : 
    (P,Q : Type) -> ConjunctionDistr -> ListTable2 P Q -> {s : P => Q => Str} = 
    \_,_,or,xs ->
    {s = \\p,q => or.s1++ xs.s1 ! p ! q ++ or.s2 ++ xs.s2 ! p ! q} ;

  comma = "," ;

} ;

--------------------------------------------------


--2 Common Nouns
--
-- Simple common nouns are defined as the type $CommNoun$ in $morpho.Deu.gf$.

--3 Common noun phrases

-- To the common nouns of morphology,
-- we add natural gender (human/nonhuman) which is needed in syntactic
-- combinations (e.g. "man who runs" - "program which runs").

oper
  CommNoun = CommonNoun ** {g : Gend} ;

  CommNounPhrase = CommNoun ;

  noun2CommNounPhrase : CommNoun -> CommNounPhrase = \man ->
    man ;

  cnGen : CommonNoun -> Gend -> CommNoun = \cn,g ->
    cn ** {g = g} ;

  cnHum : CommonNoun -> CommNoun = \cn ->
    cnGen cn Hum ;
  cnNoHum : CommonNoun -> CommNoun = \cn ->
    cnGen cn NoHum ;

--2 Noun phrases
--
-- The worst case is pronouns, which have inflection in the possessive forms. 
-- Proper names are a special case.

  NounPhrase : Type = Pronoun ;

  nameNounPhrase : ProperName -> NounPhrase = \john -> 
    {s = \\c => john.s ! toCase c ; n = Sg ; p = P3} ;

--2 Determiners
--
-- Determiners are inflected according to the nouns they determine.
-- The determiner is not inflected.
  Determiner : Type = {s : Str ; n : Num} ;


  detNounPhrase : Determiner -> CommNounPhrase -> NounPhrase = \every, man -> 
    {s = \\c => every.s ++ man.s ! every.n ! toCase c ; 
     n = every.n ; 
     p = P3
    } ;

  mkDeterminer : Num -> Str -> Determiner = \n,det -> 
    {s = det ; 
     n = n
    } ;

  everyDet = mkDeterminer Sg "every" ;
  allDet   = mkDeterminer Pl "all" ;
  mostDet  = mkDeterminer Pl "most" ;
  aDet     = mkDeterminer Sg artIndef ;
  plDet    = mkDeterminer Pl [] ;
  theSgDet = mkDeterminer Sg "the" ;
  thePlDet = mkDeterminer Pl "the" ;
  anySgDet = mkDeterminer Sg "any" ;
  anyPlDet = mkDeterminer Pl "any" ;

  whichSgDet = mkDeterminer Sg "which" ;
  whichPlDet = mkDeterminer Pl "which" ;

  whichDet = whichSgDet ; --- API

  indefNoun : Num -> CommNoun -> Str = \n,man -> 
    (indefNounPhrase n man).s ! NomP ;

  indefNounPhrase : Num -> CommNounPhrase -> NounPhrase = \n,man -> 
    {s = \\c => case n of {
                       Sg => artIndef ++ man.s ! n ! toCase c ; 
                       Pl => man.s ! n ! toCase c
                       } ;
     n = n ; p = P3
    } ;

  defNounPhrase : Num -> CommNounPhrase -> NounPhrase = \n,car -> 
    {s = \\c => artDef ++ car.s ! n ! toCase c ; n = n ; p = P3} ;

-- Genitives of noun phrases can be used like determiners, to build noun phrases.
-- The number argument makes the difference between "my house" - "my houses".
--
-- We have the variation "the car of John / the car of John's / John's car"

  npGenDet : Num -> NounPhrase -> CommNounPhrase -> NounPhrase = 
    \n,john,car -> 
      {s = \\c => variants {
             artDef ++ car.s ! n ! Nom ++ "of" ++ john.s ! GenSP ; 
             john.s ! GenP ++ car.s ! n ! toCase c
             } ;
       n = n ; 
       p = P3
      } ;

-- *Bare plural noun phrases* like "men", "good cars", are built without a 
-- determiner word.

  plurDet : CommNounPhrase -> NounPhrase = \cn -> 
    {s = \\c => cn.s ! plural ! toCase c ; 
     p = P3 ; 
     n = Pl
    } ;


--2 Adjectives
--
-- Adjectival phrases have a parameter $p$ telling if they are prefixed ($True$) or 
-- postfixed (complex APs). 

  AdjPhrase : Type = Adjective ** {p : Bool} ;

  adj2adjPhrase : Adjective -> AdjPhrase = \new -> new ** {p = True} ;

  simpleAdjPhrase : Str -> AdjPhrase = \French ->
    adj2adjPhrase (simpleAdj French) ;

--3 Comparison adjectives
--
-- Each of the comparison forms has a characteristic use:
--
-- Positive forms are used alone, as adjectival phrases ("big").

  positAdjPhrase : AdjDegr -> AdjPhrase = \big -> 
    adj2adjPhrase (ss (big.s ! Pos)) ;

-- Comparative forms are used with an object of comparison, as
-- adjectival phrases ("bigger then you").

  comparAdjPhrase : AdjDegr -> NounPhrase -> AdjPhrase = \big, you ->
	    {s = big.s ! Comp ++ "than" ++ you.s ! NomP ;
			      p = False
    } ;

-- Superlative forms are used with a modified noun, picking out the
-- maximal representative of a domain ("the biggest house").

  superlNounPhrase : AdjDegr -> CommNoun -> NounPhrase = \big, house ->
    {s = \\c => "the" ++ big.s ! Sup ++ house.s ! Sg ! toCase c ; 
     n = Sg ; 
     p = P3
    } ;


--3 Two-place adjectives
--
-- A two-place adjective is an adjective with a preposition used before
-- the complement.

  Preposition = Str ;

  AdjCompl = Adjective ** {s2 : Preposition} ;

  complAdj : AdjCompl -> NounPhrase -> AdjPhrase = \related,john ->
    {s = related.s ++ related.s2 ++ john.s ! AccP ; 
     p = False
    } ;


--3 Modification of common nouns
--
-- The two main functions of adjective are in predication ("John is old")
-- and in modification ("an old man"). Predication will be defined
-- later, in the chapter on verbs.
--
-- Modification must pay attention to pre- and post-noun
-- adjectives: "big car"/"car bigger than X"

  modCommNounPhrase : AdjPhrase -> CommNounPhrase -> CommNounPhrase = \big, car -> 
    {s = \\n => if_then_else (Case => Str) big.p 
           (\\c => big.s ++ car.s ! n ! c)
           (table {Nom => car.s ! n ! Nom ++ big.s ; Gen => variants {}}) ;
     g = car.g
    } ;


--2 Individual-valued functions

-- An individual-valued function is a common noun together with the
-- preposition prefixed to its argument ("the mother of x").
-- The situation is analogous to two-place adjectives and transitive verbs.

  Function = CommonNoun ** {s2 : Preposition} ;
--
-- N.B. the genitive case created in this way would not make sense, so we
-- have blocked it.

  appFun : Function -> NounPhrase -> NounPhrase = \mother,john -> 
    {s = table {
           GenP => nonExist ;
           _ => artDef ++ mother.s ! john.n ! Nom ++ mother.s2 ++ john.s ! GenSP
           } ;
     n = john.n ;
     p = P3
    } ;

-- The commonest case is the preposition  "of".

  funOf : CommNoun -> Function = \mother -> 
    {s = \\n => mother.s ! n ; 
     s2 = "of"
    } ;

  funOfReg : Str -> Function = \mother -> 
    funOf (nounReg mother) ;

--2 Verbs
--
--3 Verb phrases
--
-- Verb phrases are discontinuous: the two parts of a verb phrase are
-- (s) an inflected verb, (s2) infinitive and complement.
-- For instance: "doesn't" - "walk" ; "isn't" - "old" ; "is" - "a man"
-- There's also a parameter telling if the verb is an auxiliary:
-- this is needed in question.

  VerbPhrase = Verb ** {s2 : Num => Str ; isAux : Bool} ;

-- From the inflection table, we selecting the finite form as function 
-- of person and number:

  indicVerb : Verb -> Person -> Num -> Str = \v,p,n -> case n of {
    Sg => v.s ! Indic p ;
    Pl => v.s ! Indic P2
    } ;

-- A simple verb can be made into a verb phrase with an empty complement.
-- There are two versions, depending on if we want to negate the verb.
-- N.B. negation is *not* a function applicable to a verb phrase, since
-- double negations with "don't" are not grammatical.

  predVerb : Bool -> Verb -> VerbPhrase = \b,walk ->
    if_then_else VerbPhrase b 
      {s = walk.s ;
       s2 = \\_ => [] ; 
       isAux = False
      }
      {s  = \\v => contractNot (verbP3Do.s ! v) ; 
       s2 = \\_ => walk.s ! InfImp ;
       isAux = True
      } ;

-- Sometimes we want to extract the verb part of a verb phrase.

  verbOfPhrase : VerbPhrase -> Verb = \v -> {s = v.s} ;

-- Verb phrases can also be formed from adjectives ("is old"),
-- common nouns ("is a man"), and noun phrases ("ist John").
-- The third rule is overgenerating: "is every man" has to be ruled out
-- on semantic grounds.

  predAdjective : Bool -> Adjective -> VerbPhrase = \b,old ->
    {s = beOrNotBe b ;
     s2 = \\_ => old.s ;
     isAux = True
    } ;

  predCommNoun : Bool -> CommNoun -> VerbPhrase = \b,man ->
    {s = beOrNotBe b ;
     s2 = \\n => indefNoun n man ;
     isAux = True
    } ;

  predNounPhrase : Bool -> NounPhrase -> VerbPhrase = \b,john ->
    {s = beOrNotBe b ;
     s2 = \\_ => john.s ! NomP ;
     isAux = True
    } ;

-- We use an auxiliary giving all forms of "be".

  beOrNotBe : Bool -> (VForm => Str) = \b -> 
    if_then_else (VForm => Str) b 
      verbBe.s
      (table {
        InfImp   => contractNot "do" ++ "be" ;
        Indic P1 => "am" ++ "not" ;
        v        => contractNot (verbBe.s ! v)
        }) ;

--3 Transitive verbs
--
-- Transitive verbs are verbs with a preposition for the complement,
-- in analogy with two-place adjectives and functions.
-- One might prefer to use the term "2-place verb", since
-- "transitive" traditionally means that the inherent preposition is empty.
-- Such a verb is one with a *direct object*.
--
-- We also introduce verb particles only here, since for one-place verbs they
-- can as well be attached to the verb. With two-place verbs show have free
-- variation: before or after the complement 
-- ("I switch on the TV" / "I switch the TV on").

  Particle : Type = Str ;
  TransVerb : Type = Verb ** {s2 : Particle ; s3 : Preposition} ;

-- The rule for using transitive verbs is the complementization rule:

  complTransVerb : Bool -> TransVerb -> NounPhrase -> VerbPhrase = 
    \b,lookat,john ->
    let {lookatjohn = bothWays lookat.s2 (lookat.s3 ++ john.s ! AccP)} in
    if_then_else VerbPhrase b 
      {s = lookat.s ; 
       s2 = \\_ => lookatjohn ; 
       isAux = False}
      {s = \\v => contractNot (verbP3Do.s ! v) ; 
       s2 = \\_ => lookat.s ! InfImp ++ lookatjohn ;
       isAux = True} ;


-- Verbs that take direct object and a  particle:
  mkTransVerbPart : Verb -> Str -> TransVerb = \turn,off -> 
    {s = turn.s ; s2 = off ; s3 = []} ;

-- Verbs that take prepositional object, no particle:
  mkTransVerb : Verb -> Str -> TransVerb = \wait,for -> 
    {s = wait.s ; s2 = [] ; s3 = for} ;

-- Verbs that take direct object, no particle:
  mkTransVerbDir : Verb -> TransVerb = \love -> 
    mkTransVerbPart love [] ;


--2 Adverbials
--
-- Adverbials are not inflected (we ignore comparison, and treat
-- compared adverbials as separate expressions; this could be done another way).
-- We distinguish between post- and pre-verbal adverbs.

  Adverb : Type = SS ** {isPost : Bool} ;

  advPre  : Str -> Adverb = \seldom -> ss seldom ** {isPost = False} ;
  advPost : Str -> Adverb = \well   -> ss well   ** {isPost = True} ;

-- N.B. this rule generates the cyclic parsing rule $VP#2 ::= VP#2$
-- and cannot thus be parsed.

  adVerbPhrase : VerbPhrase -> Adverb -> VerbPhrase = \sings, well ->
    let {postp = orB well.isPost sings.isAux} in
    {
     s = \\v => (if_then_else Str postp [] well.s) ++ sings.s ! v ;
     s2 = \\n => sings.s2 ! n ++ (if_then_else Str postp well.s []) ;
     isAux = sings.isAux
    } ;

-- Adverbials are typically generated by prefixing prepositions.
-- The rule for creating locative noun phrases by the preposition "in"
-- is a little shaky, since other prepositions may be preferred ("on", "at").

  prepPhrase : Preposition -> NounPhrase -> Adverb = \on, it ->
    advPost (on ++ it.s ! AccP) ;

  locativeNounPhrase : NounPhrase -> Adverb = 
    prepPhrase "in" ;

-- This is a source of the "mann with a telescope" ambiguity, and may produce
-- strange things, like "cars always" (while "cars today" is OK).
-- Semantics will have to make finer distinctions among adverbials.
--
-- N.B. the genitive case created in this way would not make sense.

  advCommNounPhrase : CommNounPhrase -> Adverb -> CommNounPhrase = \car,today ->
   {s = \\n => table {
      Nom => car.s ! n ! Nom ++ today.s ; 
      Gen => nonExist
      } ;
    g = car.g
   } ;


--2 Sentences
--
-- Sentences are not inflected in this fragment of English without tense.

  Sentence : Type = SS ;

-- This is the traditional $S -> NP VP$ rule. It takes care of
-- agreement between subject and verb. Recall that the VP may already
-- contain negation. 

  predVerbPhrase : NounPhrase -> VerbPhrase -> Sentence = \john,walks ->
    ss (john.s ! NomP ++ indicVerb (verbOfPhrase walks) john.p john.n ++ 
        walks.s2 ! john.n) ;


-- This is a macro for simultaneous predication and complementization.

  predTransVerb : Bool -> NounPhrase -> TransVerb -> NounPhrase -> Sentence = 
    \b,you,see,john -> 
    predVerbPhrase you (complTransVerb b see john) ;


--3 Sentence-complement verbs
--
-- Sentence-complement verbs take sentences as complements.

  SentenceVerb : Type = Verb ;

-- To generate "says that John walks" / "doesn't say that John walks":

  complSentVerb : Bool -> SentenceVerb -> Sentence -> VerbPhrase = 
    \b,say,johnruns ->
    let {thatjohnruns = optStr "that" ++ johnruns.s} in
    if_then_else VerbPhrase b 
      {s  = say.s ; 
       s2 = \\_ => thatjohnruns ; 
       isAux = False}
      {s  = \\v => contractNot (verbP3Do.s ! v) ; 
       s2 = \\_ => say.s ! InfImp ++ thatjohnruns ;
       isAux = True} ;


--2 Sentences missing noun phrases
--
-- This is one instance of Gazdar's *slash categories*, corresponding to his
-- $S/NP$.
-- We cannot have - nor would we want to have - a productive slash-category former.
-- Perhaps a handful more will be needed.
--
-- Notice that the slash category has a similar relation to sentences as
-- transitive verbs have to verbs: it's like a *sentence taking a complement*.
-- However, we need something more to distinguish its use in direct questions:
-- not just "you see" but ("whom") "do you see".
--
-- The particle always follows the verb, but the preposition can fly:
-- "whom you make it up with" / "with whom you make it up".

  SentenceSlashNounPhrase = {s : Bool => Str ; s2 : Preposition} ;

  slashTransVerb : Bool -> NounPhrase -> TransVerb -> SentenceSlashNounPhrase = 
    \b,You,lookat ->
    let {you   = You.s ! NomP ;
         looks = indicVerb {s = lookat.s} You.p You.n ;
         look  = lookat.s ! InfImp ;
         do    = indicVerb verbP3Do You.p You.n ;
         dont  = contractNot do ;
         up    = lookat.s2
        } in
    {s  = table {
            True  => if_then_else Str b do dont ++ you ++ look ++ up ; 
            False => you ++ if_then_else Str b looks (dont ++ look) ++ up
            } ;
     s2 = lookat.s3   
    } ;


--2 Relative pronouns and relative clauses
--
-- As described in $types.Eng.gf$, relative pronouns are inflected in 
-- gender (human/nonhuman), number, and case.
--
-- We get the simple relative pronoun ("who"/"which"/"whom"/"whose"/"that"/$""$)
-- from $morpho.Eng.gf$.

  identRelPron : RelPron = relPron ;

  funRelPron : Function -> RelPron -> RelPron = \mother,which -> 
    {s = \\g,n,c => "the" ++ mother.s ! n ! Nom ++ 
                    mother.s2 ++ which.s ! g ! n ! GenSP
    } ;

-- Relative clauses can be formed from both verb phrases ("who walks") and
-- slash expressions ("whom you see", "on which you sit" / "that you sit on"). 

  RelClause : Type = {s : Gend => Num => Str} ;

  relVerbPhrase : RelPron -> VerbPhrase -> RelClause = \who,walks ->
    {s = \\g, n => who.s ! g ! n ! NomP ++ 
                   indicVerb (verbOfPhrase walks) P3 n ++ walks.s2 ! n
    } ;

  relSlash : RelPron -> SentenceSlashNounPhrase -> RelClause = \who,yousee ->
    {s = \\g,n => 
           let {youSee = yousee.s ! False} in
           variants {
             who.s ! g ! n ! AccP ++ youSee ++ yousee.s2 ;
             yousee.s2 ++ who.s ! g ! n ! GenSP ++ youSee
             }
    } ;

-- A 'degenerate' relative clause is the one often used in mathematics, e.g.
-- "number x such that x is even".

  relSuch : Sentence -> RelClause = \A ->
    {s = \\_,_ => "such" ++ "that" ++ A.s} ;

-- The main use of relative clauses is to modify common nouns.
-- The result is a common noun, out of which noun phrases can be formed
-- by determiners. No comma is used before these relative clause.

  modRelClause : CommNounPhrase -> RelClause -> CommNounPhrase = \man,whoruns ->
    {s = \\n,c => man.s ! n ! c ++ whoruns.s ! man.g ! n ;
     g = man.g
    } ;


--2 Interrogative pronouns
--
-- If relative pronouns are adjective-like, interrogative pronouns are
-- noun-phrase-like. 

  IntPron : Type = {s : CaseP => Str ; n : Num} ; 

-- In analogy with relative pronouns, we have a rule for applying a function
-- to a relative pronoun to create a new one. 

  funIntPron : Function -> IntPron -> IntPron = \mother,which -> 
    {s = \\c => "the" ++ mother.s ! which.n ! Nom ++ mother.s2 ++ which.s ! GenSP ;
     n = which.n
    } ;

-- There is a variety of simple interrogative pronouns:
-- "which house", "who", "what".

  nounIntPron : Num -> CommNounPhrase -> IntPron = \n, car ->
    {s = \\c => "which" ++ car.s ! n ! toCase c ; 
     n = n
    } ; 

  intPronWho : Num -> IntPron = \num -> {
    s = table {
      NomP  => "who" ;
      AccP  => variants {"who" ; "whom"} ;
      GenP  => "whose" ;
      GenSP => "whom"
      } ;
    n = num
  } ;

  intPronWhat : Num -> IntPron = \num -> {
    s = table {
      GenP  => "what's" ;
      _ => "what"
      } ;
    n = num
  } ;


--2 Utterances

-- By utterances we mean whole phrases, such as 
-- 'can be used as moves in a language game': indicatives, questions, imperative,
-- and one-word utterances. The rules are far from complete.
--
-- N.B. we have not included rules for texts, which we find we cannot say much
-- about on this level. In semantically rich GF grammars, texts, dialogues, etc, 
-- will of course play an important role as categories not reducible to utterances.
-- An example is proof texts, whose semantics show a dependence between premises
-- and conclusions. Another example is intersentential anaphora.

  Utterance = SS ;
  
  indicUtt : Sentence -> Utterance = \x -> ss (x.s ++ ".") ;
  interrogUtt : Question -> Utterance = \x -> ss (x.s ! DirQ ++ "?") ;


--2 Questions
--
-- Questions are either direct ("are you happy") or indirect 
-- ("if/whether you are happy").

param 
  QuestForm = DirQ | IndirQ ;

oper
  Question = SS1 QuestForm ;

--3 Yes-no questions 
--
-- Yes-no questions are used both independently 
-- ("does John walk" / "if John walks")
-- and after interrogative adverbials 
-- ("why does John walk" / "why John walks").
-- 
-- It is economical to handle with all these cases by the one
-- rule, $questVerbPhrase'$. The word ("ob" / "whether") never appears
-- if there is an adverbial.

  questVerbPhrase : NounPhrase -> VerbPhrase -> Question = 
    questVerbPhrase' False ;

  questVerbPhrase' : Bool -> NounPhrase -> VerbPhrase -> Question = 
    \adv,john,walk ->
    {s = table {
      DirQ   => if_then_else Str walk.isAux 
                  (indicVerb (verbOfPhrase walk) john.p john.n ++
                   john.s ! NomP ++ walk.s2 ! john.n) 
                  (indicVerb verbP3Do john.p john.n ++ 
                   john.s ! NomP ++ walk.s ! InfImp ++ walk.s2 ! john.n) ;
      IndirQ => if_then_else Str adv [] (variants {"if" ; "whether"}) ++ 
                (predVerbPhrase john walk).s
      }
    } ;



--3 Wh-questions
--
-- Wh-questions are of two kinds: ones that are like $NP - VP$ sentences,
-- others that are line $S/NP - NP$ sentences.

  intVerbPhrase : IntPron -> VerbPhrase -> Question = \who,walk ->
    {s = \\_ => who.s ! NomP ++ indicVerb (verbOfPhrase walk) P3 who.n ++ 
                walk.s2 ! who.n 
    } ;

  intSlash : IntPron -> SentenceSlashNounPhrase -> Question = \who,yousee ->
    {s = \\q =>
           let {youSee = case q of {
                  DirQ   => yousee.s ! True ; 
                  IndirQ => yousee.s ! False
                  }
           } in
           variants {
             who.s ! AccP ++ youSee ++ yousee.s2 ;
             yousee.s2 ++ who.s ! GenSP ++ youSee
             } 
    } ;

--3 Interrogative adverbials
--
-- These adverbials will be defined in the lexicon: they include
-- "when", "where", "how", "why", etc, which are all invariant one-word
-- expressions. In addition, they can be formed by adding prepositions
-- to interrogative pronouns, in the same way as adverbials are formed
-- from noun phrases. 

  IntAdverb = SS ;

  prepIntAdverb : Preposition -> IntPron -> IntAdverb = \at, whom ->
    ss (at ++ whom.s ! AccP) ;

-- A question adverbial can be applied to anything, and whether this makes
-- sense is a semantic question.

  questAdverbial : IntAdverb -> NounPhrase -> VerbPhrase -> Question = 
    \why, you, walk ->
    {s = \\q => why.s ++ (questVerbPhrase' True you walk).s ! q} ;


--2 Imperatives
--
-- We only consider second-person imperatives. 

  Imperative = SS1 Num ;

  imperVerbPhrase : VerbPhrase -> Imperative = \walk -> 
    {s = \\n => walk.s ! InfImp ++ walk.s2 ! n} ;

  imperUtterance : Num -> Imperative -> Utterance = \n,I ->
    ss (I.s ! n ++ "!") ;


--2 Coordination
--
-- Coordination is to some extent orthogonal to the rest of syntax, and
-- has been treated in a generic way in the module $CO$ in the file
-- $coordination.gf$. The overall structure is independent of category,
-- but there can be differences in parameter dependencies.
--
--3 Conjunctions
--
-- Coordinated phrases are built by using conjunctions, which are either
-- simple ("and", "or") or distributed ("both - and", "either - or").

  Conjunction = CO.Conjunction ;
  ConjunctionDistr = CO.ConjunctionDistr ;

--3 Coordinating sentences
--
-- We need a category of lists of sentences. It is a discontinuous
-- category, the parts corresponding to 'init' and 'last' segments
-- (rather than 'head' and 'tail', because we have to keep track of the slot between
-- the last two elements of the list). A list has at least two elements.

  ListSentence : Type = SD2 ;

  twoSentence : (_,_ : Sentence) -> ListSentence = CO.twoSS ;

  consSentence : ListSentence -> Sentence -> ListSentence =
    CO.consSS CO.comma ;

-- To coordinate a list of sentences by a simple conjunction, we place
-- it between the last two elements; commas are put in the other slots,
-- e.g. "du rauchst, er trinkt und ich esse".

  conjunctSentence : Conjunction -> ListSentence -> Sentence = \c,xs ->
    ss (CO.conjunctX c xs) ;

-- To coordinate a list of sentences by a distributed conjunction, we place
-- the first part (e.g. "either") in front of the first element, the second
-- part ("or") between the last two elements, and commas in the other slots.
-- For sentences this is really not used.

  conjunctDistrSentence : ConjunctionDistr -> ListSentence -> Sentence = 
    \c,xs ->
    ss (CO.conjunctDistrX c xs) ;

--3 Coordinating adjective phrases
--
-- The structure is the same as for sentences. The result is a prefix adjective
-- if and only if all elements are prefix.

  ListAdjPhrase : Type = SD2 ** {p : Bool} ;

  twoAdjPhrase : (_,_ : AdjPhrase) -> ListAdjPhrase = \x,y ->
    CO.twoStr x.s y.s ** {p = andB x.p y.p} ;

  consAdjPhrase : ListAdjPhrase -> AdjPhrase -> ListAdjPhrase =  \xs,x ->
    CO.consStr CO.comma xs x.s ** {p = andB xs.p x.p} ;

  conjunctAdjPhrase : Conjunction -> ListAdjPhrase -> AdjPhrase = \c,xs ->
    ss (CO.conjunctX c xs) ** {p = xs.p} ;

  conjunctDistrAdjPhrase : ConjunctionDistr -> ListAdjPhrase -> AdjPhrase = 
    \c,xs ->
    ss (CO.conjunctDistrX c xs) ** {p = xs.p} ;

--2 Subjunction
--
-- Subjunctions ("when", "if", etc) 
-- are a different way to combine sentences than conjunctions.
-- The main clause can be a sentences, an imperatives, or a question,
-- but the subjoined clause must be a sentence.
--
-- There are uniformly two variant word orders, e.g. 
-- "if you smoke I get angry"
-- and "I get angry if you smoke".

  Subjunction = SS ;

  subjunctSentence : Subjunction -> Sentence -> Sentence -> Sentence = 
    \if, A, B -> 
    ss (subjunctVariants if A.s B.s) ;

  subjunctImperative : Subjunction -> Sentence -> Imperative -> Imperative = 
    \if, A, B -> 
    {s = \\n => subjunctVariants if A.s (B.s ! n)} ;

  subjunctQuestion : Subjunction -> Sentence -> Question -> Question = 
    \if, A, B ->
    {s = \\q => subjunctVariants if A.s (B.s ! q)} ;

  subjunctVariants : Subjunction -> Str -> Str -> Str = \if,A,B ->
    variants {if.s ++ A ++ "," ++ B ; B ++ "," ++ if.s ++ A} ;


--2 One-word utterances
-- 
-- An utterance can consist of one phrase of almost any category, 
-- the limiting case being one-word utterances. These
-- utterances are often (but not always) in what can be called the
-- default form of a category, e.g. the nominative.
-- This list is far from exhaustive.

  useNounPhrase : NounPhrase -> Utterance = \john ->
    postfixSS "." (defaultNounPhrase john) ;

  useCommonNounPhrase : Num -> CommNounPhrase -> Utterance = \n,car -> 
    useNounPhrase (indefNounPhrase n car) ;

  useRegularName : SS -> NounPhrase = \john -> 
    nameNounPhrase (nameReg john.s) ;

-- Here are some default forms.

  defaultNounPhrase : NounPhrase -> SS = \john -> 
    ss (john.s ! NomP) ;

  defaultQuestion : Question -> SS = \whoareyou ->
    ss (whoareyou.s ! DirQ) ;

  defaultSentence : Sentence -> Utterance = \x -> 
    x ;



---------------------------------------



--
-- general NL
--

-- the gentive form here is perhaps not obvious?
oper conjoinNP : Str -> (n1,n2 : NounPhrase) -> NounPhrase = \and, n1, n2 -> 
		{s = \\c => n1.s!c ++ and ++ n2.s!c; n = Pl; p = P3};
		
	appFunSg : Function -> NounPhrase -> NounPhrase = \mother,john -> 
	    {s = table {
		GenP => nonExist ;
		_  => artDef ++ mother.s ! Sg ! Nom ++ mother.s2 ++ john.s ! GenSP };
     n = Sg ;
     p = P3
    } ;
    
  funX : CommNoun -> Str -> Function = \mother,prep -> 
    {s = \\n => mother.s ! n  ; s2 = prep} ;
  funXReg : Str -> Str -> Function = \mother,prep -> funX (nounReg mother) prep;

  -- for handling ind.valued fuctions with "postfix notation" (e.g. the string converted to ...)
  postfixIndFun : Str -> NounPhrase -> NounPhrase = \f, np ->  { s = \\c =>
		np.s ! c ++ f;
		n=np.n;
		p = np.p
  };
		
  constant : Str -> NounPhrase = \c -> nameNounPhrase (nameReg c);
  numeral : Str -> NounPhrase = constant;



--
-- Classes
--

lintype ClassL  = CommNoun ** {s42 : Str}; -- let's hope CN has less than 42 "s"-field
lincat Class = ClassL;
oper class2id : ClassL -> Str = \c -> c.s42;
	class2cn: ClassL -> CommNoun = \c -> {s = c.s; g = c.g};

oper mkClassNoHum : Str -> Str -> ClassL = \noun, classId -> 
		(cnNoHum (nounReg noun)) ** {s42 = classId};
	mkClassHum : Str -> Str -> ClassL = \noun, classId -> 
		(cnHum (nounReg noun)) ** {s42 = classId};
	mkClass = mkClassNoHum;

--
-- Subtyping 
--
lin coerce sub super proof obj = obj;
    subTypeTrans a b c pab pbc= {s = ["every"] ++ a.s!Sg!Nom ++ ["is a"] ++ c.s!Sg!Nom ++ "," 
    	++ "since" ++ pab.s ++ "and" ++ pbc.s}; 
   subTypeRefl a = {s = "a" ++ a.s!Sg!Nom ++ ["is a"] ++ a.s!Sg!Nom ++ ["is a"] ++ a.s!Sg!Nom};  

printname coerce = "[coercion]";



--
-- Instances
--

lintype Inst = NounPhrase; 
lincat Instance = Inst;
oper test : Str -> Inst = \t -> {s = \\c => t; n=Sg; p=P3};
lindef Instance = test; -- \t -> {s = \\c => t; n=Sg; p=P3};
oper 
	np2inst : NounPhrase -> Inst = \np -> np;    
	inst2np : Inst -> NounPhrase = \inst -> inst;

-- should we have "True" as the first argument to predAdjective?
oper equalTo : Inst -> Inst -> Sentence = \a,b ->
	(predVerbPhrase a (predAdjective True (complAdj {s="equal"; s2="to"} b)));
     notEqualTo : Inst -> Inst -> Sentence = \a,b ->
	(predVerbPhrase a (predAdjective False (complAdj {s=["equal"]; s2="to"} b)));
    littleAdj : AdjDegr = mkAdjDegr "little" "less" "least";
    greatAdj : AdjDegr = adjDegrReg "great";
     lessThan : Inst -> Inst -> Sentence = \a,b -> -- a is less than b
	      predVerbPhrase a (predAdjective True (comparAdjPhrase littleAdj b));
     greaterThan : Inst -> Inst -> Sentence = \a, b -> -- a is greater than b
	      predVerbPhrase a (predAdjective True (comparAdjPhrase greatAdj b));
     
     instFunOf : Str -> Inst -> Inst = \size, object ->
	       np2inst (appFunSg (funOfReg size) (inst2np object));
     instInfix : Str -> Inst -> Inst -> Inst = \plus, a, b ->
	       np2inst (conjoinNP plus (inst2np a) (inst2np b));	       
     instPrefix : Str -> Inst -> Inst = \not, a -> np2inst { 
		s = \\c => not ++ (inst2np a).s!c;
		n = (inst2np a).n;
		p = (inst2np a).p};
	       
    conjoinInst : Str -> Inst -> Inst -> Inst = \and, a, b ->
		np2inst (conjoinNP and (inst2np a) (inst2np b));
	      
	


lin atPre _ i = {s = \\c => i.s!c ++ ["at the start of the operation"]; n=i.n; p=i.p};

--
-- Attributes / queries (Features)
--

oper mkAttr : Inst -> Str -> Inst = \obj, attr ->
	np2inst(appFunSg (funOfReg attr) obj);

     -- handle attributes like "isMarried"
     mkIsAttr : Inst -> Str -> Sentence = \obj, married ->
	       (predVerbPhrase (inst2np obj) (predAdjective True (ss married)));
     -- arbitrary boolean-valued attributes
     -- the property prop holds for obj
     mkSentAttr : Inst -> Str -> Sentence = \obj, prop ->
		(predVerbPhrase (constant (["the property"] ++ prop))
		(complTransVerb True (mkTransVerb (regVerbP3 "hold") "for") obj));

-- build list of arguments/parameters to queries:	
lintype  ArgList = {s : Str; empty : Bool};
oper argNil : ArgList = {s = ""; empty = True};
	argCons : Inst -> ArgList -> ArgList = 
		\x, xs -> {s = case xs.empty of {
			True => (inst2np x).s!NomP;
			False => xs.s ++ "," ++ (inst2np x).s!NomP
		}; empty = False};

	mkQuery : Inst -> Str -> ArgList -> Inst = \obj, query, args -> 
		np2inst	(appFunSg (funOfReg (query ++ "(" ++ args.s ++ ")")) obj);
     -- handle queries like "isMarried()"
     mkIsQuery : Inst -> Str -> Sentence = \obj, married ->
	       (predVerbPhrase (inst2np obj) (predAdjective True (ss married)));
     -- arbitrary boolean-valued queries
     -- the property prop(x,y,z) holds for obj
     mkSentQuery : Inst -> Str -> ArgList -> Sentence = \obj, query, args ->
		(predVerbPhrase (constant (["the property"] ++ query ++ "(" ++ args.s ++ ")"))
		(complTransVerb True (mkTransVerb (regVerbP3 "hold") "for") obj));

-- associations
   mkAssoc : Str -> Inst -> Inst = \thing, obj -> -- the set of the things associated with obj
	   postfixIndFun (["associated with"] ++ (inst2np obj).s!AccP)
	    (np2inst ((appFunSg (funOfReg "set") (defNounPhrase Pl (nounReg thing)))));
			

--
-- Sentences 
--
lincat Sent = Sentence;

oper s2sent : Str -> Sentence =  \str -> {s = str};
lin	bool2sent b = equalTo b (constant "true");
	sent2bool s = (np2inst (appFunSg (funOfReg ["truth value of the condition that"]) 
		(constant s.s)));

printname sent2bool = "[use a sentence as a boolean]";
	  bool2sent = "[use a boolean as a sentence]";


lin atPreSent s = s2sent (s.s ++ ["at the start of the operation"]);

--
-- self & result
--
lin self c _ = np2inst (detNounPhrase theSgDet c);
lin result _ _ = np2inst (detNounPhrase theSgDet (nounReg "result"));


--
-- pre- and postconditions
--
oper beginBulletList : Str = ""; -- "\\begin{itemize}";
	 bullet : Str = "(*)"; --"\\item";
	endBulletList : Str = ""; --"\\end{itemize}";
	wrapBullets : Str -> Str = \str -> beginBulletList ++ str ++ endBulletList;

param ListSize = LSempty | LSone | LSmany;
oper LSsucc : ListSize => ListSize = table {LSempty => LSone; LSone => LSmany; LSmany => LSmany};
	LS2Bool : ListSize => Bool = table {LSempty => False; _ => True};


lintype ConstraintBodyL = {s0, s1, s2 : Str; invC, preC, postC : ListSize};
lincat ConstraintBody = ConstraintBodyL;   -- one string for inv, pre and post, respectively


lin
	emptyC = {s0,s1,s2 = ""; invC, preC, postC = LSempty};
	invC i cs = {s0 = bullet ++ i.s ++ cs.s0; invC = LSsucc ! cs.invC;
		s1 = cs.s1; preC = cs.preC;
		s2 = cs.s2; postC = cs.postC};
	preC p cs = {s0 = cs.s0 ; 
		s1 = bullet ++ p.s ++ cs.s1; 
		s2 = cs.s2; invC = cs.invC; preC = LSsucc ! cs.preC; postC = cs.postC};
	postC p cs = {s0 = cs.s0; s1 = cs.s1; invC = cs.invC; preC = cs.preC;
		s2 = bullet ++ p.s ++ cs.s2; postC = LSsucc ! cs.postC};

	invHC s = s;

oper 
	LSand1 : (l1,l2,l3 : ListSize) -> Str = \l1, l2, l3 -> -- "and" between inv and pre?
		LSand1' (LS2Bool ! l1) (LS2Bool ! l2) (LS2Bool ! l3);
	LSand1' : (b1,b2,b2 : Bool) -> Str = \b1, b2, b3 ->
		if_then_else Str (orB (andB b1 b2) (andB b1 b3)) "and" "";
	LSand2 : (l2,l3 : ListSize) -> Str = \l2, l3 ->    -- "then" between pre and post?
		LSand2' (LS2Bool ! l2) (LS2Bool ! l3);
	LSand2' : (b2,b3 : Bool) -> Str = \b2, b3 ->
		if_then_else Str (andB b2 b3) "then" "";

oper cb2str : ConstraintBodyL -> Str = 
	\cb -> (case cb.invC of {
		 LSempty => "";
		 LSone => ["this invariant should hold :"] ++ wrapBullets cb.s0;
		 LSmany => ["these invariants should hold :"] ++ wrapBullets cb.s0
	}) ++ LSand1 cb.invC cb.preC cb.postC ++ 
	(case cb.preC of {
		LSempty => "";
		LSone => ["given this precondition :"] ++ wrapBullets cb.s1;
		LSmany => ["given these preconditions :"] ++ wrapBullets cb.s1
	}) ++ LSand2 cb.preC cb.postC ++ 
	(case cb.postC of {
		LSempty => "";
		LSone => ["this postcondition should hold :"] ++ wrapBullets cb.s2;
		LSmany => ["these postconditions should hold : "] ++ wrapBullets cb.s2
	});


	strip2varCB : {v0 ,v1 : Str} ** ConstraintBodyL  -> ConstraintBodyL = \cb -> 
		{s0=cb.s0; s1=cb.s1; s2=cb.s2; invC=cb.invC; preC=cb.preC; postC=cb.postC};

--
-- constraint
--

lintype ConstraintL = {s:Str};
lincat Constraint = ConstraintL;

oper contextOp00 : Str -> Str -> ConstraintBodyL -> ConstraintL = \class, op, body ->
	ss (["for the operation"] ++ op ++ ["of the class"] ++ class ++ ["we have that"] ++ cb2str body);
	contextOp10 : (class, op, args : Str) -> ConstraintBodyL -> ConstraintL =
		\class, op, args, body -> ss (["for the operation"] ++ op ++ "(" ++ args ++ ")" ++ 
			["of the class"] ++ class ++ ["we have that"] ++ cb2str body);
	contextOp01 : (class, op,ret : Str) -> ConstraintBodyL -> ConstraintL = 
		\class, op, ret, body -> ss (["for the operation"] ++ op
		++ "()" ++ ":" ++ ret  ++ ["of the class"] ++ class ++ ["we have that"] ++ cb2str body);
	contextOp11 : (class, op, args, ret : Str) -> ConstraintBodyL -> ConstraintL = 
		\class, op, args, ret, body -> ss (["for the operation"] ++ op
		++ "(" ++ args ++ ")" ++ ":" ++ ret  ++ ["of the class"] ++ class ++ ["we have that"] ++ cb2str body);

	contextClass : Str -> ConstraintBodyL -> ConstraintL = 
		\class, body -> ss (["for the class"] ++ class ++ ["we have that"] ++ cb2str body);		


--
-- let expressions
--

lincat LetExp = Inst ** {s10 : Str; listSize : ENumber}; -- Inst + the list of bindings

lin letExp2Inst _ le = np2inst {
	s = \\c => "define" ++ le.s10 ++ "in" ++ le.s!c;
	n = le.n; p = le.p
	};

lin letSingleAttr _ _ defn body = {
		s = body.s; n = body.n; p = body.p; 
		s10 = body.v ++ "as" ++ defn.s ! NomP;
		listSize = E1};	
	letConsAttr _ _ defn body = {
		s = body.s; n = body.n; p = body.p; 
		s10 = case body.listSize of {
			E1 => body.v ++ "as" ++ defn.s!NomP ++ "and" ++ body.s10;
			_ => body.v ++ "as" ++ defn.s!NomP ++ "," ++ body.s10
		};
		listSize = eNext body.listSize
	};
	
lincat LetExpSent = Sentence ** {s10 : Str; listSize : ENumber}; -- Sent + the list of bindings

lin letExp2Sent le = s2sent ("define" ++ le.s10 ++ "in" ++ le.s);

lin letSingleAttrS _ defn body = {
		s = body.s; 
		s10 = body.v ++ "as" ++ defn.s ! NomP;
		listSize = E1};	
	letConsAttrS _ defn body = {
		s = body.s; 
		s10 = case body.listSize of {
			E1 => body.v ++ "as" ++ defn.s!NomP ++ "and" ++ body.s10;
			_ => body.v ++ "as" ++ defn.s!NomP ++ "," ++ body.s10
		};
		listSize = eNext body.listSize
	};





-----------------------------------


--
-- OclType
--

lin OclType = mkClassNoHum ["OCL type"] "OclType";

    class2oclType c = constant (class2id c);

    otName t = np2inst (appFunSg (funOfReg "name") t);
    otAttributes t = np2inst (appFunSg (funOfReg ["set of the attributes"]) t);
    otOperations t = np2inst (appFunSg (funOfReg ["set of the operations"]) t);
    otSupertypes t = np2inst (appFunSg (funOfReg ["set of the direct supertypes"]) t);
    otAllSupertypes t = np2inst (appFunSg (funOfReg ["set of all supertypes"]) t);
    otAllInstancesStrict c t = np2inst (appFunSg (funOfReg ["set of all instances"]) t);

--
-- OclAny
--
lin OclAny = mkClassNoHum "object" "OclAny";

    conforms2OclAny c = ss (class2id c ++ ["is a subtype of OclAny"]);

    anyEq a b = s2sent ((inst2np a).s!NomP ++ ["is the same object as"]  ++ (inst2np b).s!NomP);
    anyNeq a b = s2sent ((inst2np a).s!NomP ++ ["is not the same object as"]  ++ (inst2np b).s!NomP);
    anyOclIsKindOf a t = s2sent ( ["one of the (super-)types of"] ++ (inst2np a).s!AccP ++ 
         ["is equal to"] ++ (inst2np t).s!NomP);
    anyOclIsTypeOf a t = s2sent ( ["one of the types of"] ++ (inst2np a).s!AccP ++ 
         ["is equal to"] ++ (inst2np t).s!NomP);
    anyOclAsTypeStrict a c t = postfixIndFun (["considered as an instance of"] ++ (inst2np t).s!AccP) a;
    anyOclAsTypeClass a c = postfixIndFun (["considered as"] ++ artIndef ++ c.s!Sg!Nom) a;
--    anyOclAsTypeExpr a c _ = postfixIndFun (["considered as"] ++ artIndef ++ c.s!Sg!Nom) a;
    anyOclInState a s = s2sent ((inst2np a).s!NomP ++ ["is in state"] ++ (inst2np s).s!NomP);
    anyOclIsNew a = s2sent ((inst2np a).s!NomP ++ ["is created during the performing the operation"]);

--
-- OclState
--
lin OclState = mkClassNoHum "state" "OclState";


--
-- OclExpression
--

lin OclExpression = mkClassNoHum "expression" "OclExpression";

    inst2OclExpr _ i = i;  -- circular, of course

    oeEvaluationType i = np2inst (appFunSg (funOfReg ["type of"]) i);



--
-- Integer
--
lin Integer = mkClassNoHum  "integer" "Integer";
	int i = np2inst (numeral i.s);
	intAdd = instInfix "plus"; 
	intSub = instInfix "minus";
	intUnaryMinus = instFunOf ["negative value"];
	intTimes = instInfix "times"; 
	intDiv2Real = instInfix ["divided by"]; 
	intDiv2Int = instInfix ["integer divided by"];
	intMod = instInfix "modulo";

{- 
	-- These could be added as variants:
	
	intAdd a b = np2inst (appFunSg (funOfReg "sum") (conjoinNP "and" a b));  -- the sum of a and b
	intSub a b = np2inst (appFunSg (funXReg "difference" "between") (conjoinNP "and" a b)); 
	intTimes a b = np2inst (appFunSg (funOfReg "product") (conjoinNP "and" a b));
-}

	intAbs = instFunOf ["absolute value"];
	intMax a b = instFunOf "maximum" (conjoinInst "and" a b);
	intMin a b = instFunOf "minimum" (conjoinInst "and" a b);
	
	intEq = equalTo;
	intLT = lessThan;
	intGT = greaterThan;
	-- the following is not pretty, a better way?
	intLTE a b = let {ltb =complAdj {s = ["less than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	intGTE a b = let {ltb =complAdj {s = ["greater than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	
	Zero = numeral "zero";
	One = numeral "one";
	Two = numeral "two";
	
--
-- Real
--

lin
	Real = mkClassNoHum "real" "Real";
	realEq = equalTo;
	realNeq = notEqualTo;
	realAdd = instInfix "add"; 
	realSub = instInfix "sub"; 
	realTimes = instInfix "times"; 
	realDiv = instInfix ["divided by"];
	realUnaryMinus = instFunOf ["negative value"];
	realAbs = instFunOf ["absolute value"];
	realFloor a = np2inst (appFunSg (funXReg ["largest integer less than or equal"] "to") a);
	realRound a = np2inst (appFunSg (funXReg ["integer closest"] "to") a);
	realMax a b = instFunOf "maximum" (conjoinInst "and" a b);
	realMin a b = instFunOf "minimum" (conjoinInst "and" a b);
	realLT = lessThan; 
	realGT = greaterThan; 
	-- the following is not pretty, a better way?
	-- we need "or" on adjectives in resource grammar
	realLTE a b = let {ltb =complAdj {s = ["less than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	realGTE a b = let {ltb =complAdj {s = ["greater than or equal"];s2 = "to"} b} in
		 (predVerbPhrase a (predAdjective True ltb));
	 

--
-- String
--
lin StringC = mkClassNoHum "string" "String";
	stringEq a b = equalTo a b;
	stringLiteral lit = constant lit.s;
	stringSize  = instFunOf "size";
	stringConcat a b = instFunOf "concatenation" (conjoinInst "and" a b);
	stringToUpper = postfixIndFun ["converted to uppercase"];
	stringToLower = postfixIndFun ["converted to lowercase"];
	stringSubstring s a b = np2inst ( -- here we have the problems of indices...
			postfixIndFun (["ranging from index"] ++ a.s!NomP ++ ["to index"] ++ b.s!NomP) 
			(appFunSg (funOfReg "substring") s)
	);
			
	

--
-- Bool
--

lin BoolC = mkClassNoHum "bool" "Boolean";
	true = (constant "true");
	false = (constant "false");
	boolEq = equalTo;
	orBool = instInfix "or"; 
	xorBool = instInfix "xor";
	andBool = instInfix "and";
	notBool = instPrefix "not"; --np2inst (constant ("not" ++ a.s!NomP));
	impliesBool = instInfix "implies";
			
--
-- Sent 
--
lin sentEq a b = s2sent (a.s ++ ["if and only if"] ++ b.s);
-- use conjunctions
	orS a b = s2sent (a.s ++ "or" ++ b.s);
	xorS a b = s2sent ("either" ++ a.s ++ "or" ++ b.s);
	andS a b = s2sent (a.s ++ "and" ++ b.s);
	notS a = s2sent (["it is not the case that"] ++ a.s);
	impliesS a b = s2sent (a.s ++ ["implies that"] ++ b.s);
	-- ifThenElse is a bit tricky: do we really know that t 
	-- and e have the same number and person?
	-- ideally we should have a resource grammar which supports conditional expressions
 	ifThenElse _ cond t e = {
 		s = \\c => t.s!c ++ [", if"] ++ cond.s ++ "," ++ e.s!c ++ "otherwise";
 		n = t.n; p = t.p};
	ifThenElseS cond t e = s2sent ("if" ++ cond.s ++ "then" ++ t.s ++ "else" ++ e.s);



--
-- Iterators
--

-- a InstIter is an instance and a list of bound variables
-- this is not abstract, we would like to simply extend Inst with 
-- a string field for the variable list. Note that Inst = NounPhrase = Pronoun
lintype InstIterL = {s : CaseP => Str ; n : Num ; p : Person; s1 : Str};
lincat InstIter = InstIterL;

lin instIterSingle varC _ body = {
		s1 = body.v;
		s0 = body.s0; n = body.n; p = body.p};
	instIterMany varC _ body = {
		s1 = body.s1 ++ "," ++ body.v;
		s0 = body.s0; n = body.n; p = body.p};

-- a SentIter is a sentence + a list of bound variables
lintype SentIterL = {s0, s1 : Str; moreThanOne : Bool};
lincat SentIter = SentIterL;
lin	sentIterSingle varC  body = {
		s1 = body.v;
		s0 = body.s0;
		moreThanOne = False};
	sentIterMany varC body = {
		s1 = body.s1 ++ "," ++ body.v;
		s0 = body.s0;
		moreThanOne = True};
		
--
-- Collections
--

-- general operations

-- are all forms here really correct?
oper mkCollClass : Str -> Str -> ClassL-> ClassL = \collection, Collection, Integer -> 
	let {integers = Integer.s!Pl!Nom} in 
		(mkNoun (collection ++ "of" ++ integers) (collection+"s" ++ "of" ++ integers)
		(collection ++ "of" ++ integers+"'") (collection+"s" ++ "of" ++ integers+"'")
		) ** {g = NoHum; s42 = Collection ++ class2id Integer};

    collIncluding : Str -> Inst -> Inst -> Inst = \collType, coll, elem -> 
        instFunOf (collType ++ ["containing all elements"]) (instInfix "plus" coll elem);
    collExcluding : Str -> Inst -> Inst -> Inst = \collType, coll, elem -> 
        instFunOf (collType ++ ["containing all elements"]) (instInfix "without" coll elem);
    collSelect : Str -> Inst -> SentIterL -> Inst = \collType, coll, iter -> 
	-- let {satisfy = if_then_else Str	iter.moreThanOne "satisfies" "satisfy"} in 
	postfixIndFun (["which satisfy"] ++ iter.s0) 
         (instFunOf ("the" ++ collType ++ ["all elements"] ++ iter.s1) coll);
    collReject : Str -> Inst -> SentIterL -> Inst = \collType, coll, iter -> 
	postfixIndFun (["which do not satisfy"] ++ iter.s0) 
         (instFunOf ("the" ++ collType ++ ["all elements"] ++ iter.s1) coll);
    collCollect : Str -> Inst -> InstIterL -> Inst = \collType, coll, iter -> 
        -- the collType resulting from evaluating iter.s0 for all iter.s1 in coll
	-- the result of evaluating "iter.s0" for all iter.s1 in coll
	-- NomP correct for coll.s?
         postfixIndFun (["for all"] ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP) 
	 (instFunOf "result" (constant (["evaluating \""] ++ (inst2np iter).s!NomP ++ "\"")));
    collCount : Inst -> Inst -> Inst = \coll, elem ->
        np2inst 
	  (appFunSg (funXReg (["number of times"] ++ (inst2np elem).s!NomP ++ "occurs") "in") 
	  coll);
		
lin Collection c = mkCollClass "collection" "Collection" c; --mkClassNoHum (["collection of"] ++ c.s!Pl!Nom) ("Collection" ++ class2id c);

lin size _ coll = np2inst (appFunSg (funOfReg "size") coll);
	includes _ _ coll elem = predVerbPhrase (inst2np coll) 
		(complTransVerb True (mkTransVerbDir (regVerbP3 "contain")) (inst2np elem));
	excludes _ _ coll elem = predVerbPhrase (inst2np coll) 
		(complTransVerb False (mkTransVerbDir (regVerbP3 "contain")) (inst2np elem)); -- correct?
	count _ = collCount; 
	includesAll _ coll1 coll2 = predVerbPhrase (inst2np coll1) 
		(complTransVerb True (mkTransVerbDir (mkVerbP3 ["contains all elements of"] 
		["contain all elements of"])) (inst2np coll2));
	excludesAll _ coll1 coll2 = predVerbPhrase (inst2np coll1) 
		(complTransVerb True (mkTransVerbDir (mkVerbP3 ["contains none of the elements of"]
		["contain none of the elements of"])) 
		(inst2np coll2));
	isEmpty _ coll = (predVerbPhrase coll (predAdjective True (ss "empty")));
	notEmpty _ coll = (predVerbPhrase coll (predAdjective True (ss ["not empty"])));
	sum _ coll = np2inst (appFunSg (funOfReg ["the sum of all elements"]) coll);
	forAll c coll iter = s2sent (["for all"] ++ (class2cn c).s!Pl!Nom ++ iter.s1 ++
		"in" ++ (inst2np coll).s!NomP ++ "," ++ iter.s0);
	exists c coll iter = s2sent ((case iter.moreThanOne of {
		True => ["there exist"] ++ (class2cn c).s!Pl!Nom ;   
		False => ["there exists"] ++ artIndef ++ (class2cn c).s!Sg!Nom
	}) ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP ++ ["such that"] ++ iter.s0);
	isUnique c d coll iter = s2sent (iter.s!NomP ++ ["is unique for every"] ++ (class2cn c).s!Pl!Nom ++
		iter.s1 ++ "in" ++ (inst2np coll).s!NomP); -- need quotes here?
	sortedBy c d coll iter = constant (["the sequence of all"] ++ (class2cn c).s!Pl!Nom ++ iter.s1 ++
		"in" ++ (inst2np coll).s!NomP ++ ["sorted by"] ++ iter.s0!NomP);
	iterate c a coll init body = constant (["the result of iterating over every"]  ++ (class2cn c).s!Sg!Nom ++ 
		body.v0 ++ "in" ++ (inst2np coll).s!NomP ++ ["with accumulator"] ++ body.v1 ++ ["initialized to"] ++
		(inst2np init).s!NomP ++ [", in each step evaluating"] ++ body.s!NomP ); -- (inst2np body).s!NomP does not work???
	any c coll iter = constant ("any" ++ (class2cn c).s!Sg!Nom ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP ++
		["such that"] ++ iter.s0);
	one c coll iter = s2sent ((case iter.moreThanOne of {
		True => ["there exist exactly one"] ++ (class2cn c).s!Pl!Nom ;   -- should be variants{} ?
		False => ["there exists exactly one"] ++ artIndef ++ (class2cn c).s!Sg!Nom
	}) ++ iter.s1 ++ "in" ++ (inst2np coll).s!NomP ++ ["such that"] ++ iter.s0);

--
-- sets 
--

lin Set c = mkCollClass "set" "Set" c;
    unionSS _ a b = instFunOf "union" (conjoinInst "and" a b);
    unionSB _ a b = instFunOf "union" (conjoinInst "and" a b);
    setEq c = equalTo;
    intersectionSS c a b = instFunOf "intersection" (conjoinInst "and" a b);
    intersectionSB c a b = instFunOf "intersection" (conjoinInst "and" a b);
    subS c a b = instFunOf ["set difference"] (conjoinInst "minus" a b); 
    includingS c = collIncluding "set"; 
    excludingS c = collExcluding "set";
    symmetricDiff c a b = instFunOf ["symmetric difference"] (conjoinInst "and" a b);
    selectSet c = collSelect "set";   
    rejectSet c = collReject "set";
    collectSet c d = collCollect "bag";
    countSet _ = collCount;
    setAsSequence _ = postfixIndFun ["as a sequence"]; -- alternative is a circular rule "= c"
    setAsBag _ = postfixIndFun ["as a bag"];

--
-- bags
--
lin Bag c = mkCollClass "bag" "Bag" c;
    bagEq _ = equalTo;
    unionBB _ a b = instFunOf "union" (conjoinInst "and" a b);
    unionBS _ a b = instFunOf "union" (conjoinInst "and" a b);
    intersectionBB _ a b = instFunOf "intersection" (conjoinInst "and" a b);
    intersectionBS _ a b = instFunOf "intersection" (conjoinInst "and" a b);
    includingB c = collIncluding "bag";
    excludingB c = collExcluding "bag";
    selectBag c = collSelect "bag";
    rejectBag c = collReject "bag";
    collectBag _ _ = collCollect "bag";
    countBag _ = collCount;
    bagAsSequence _ = postfixIndFun ["as a sequence"];
    bagAsSet _ = postfixIndFun ["as a set"];

--
-- sequences
--
lin Sequence c = mkCollClass "sequence" "Sequence" c;
    sequenceEq _ = equalTo;
    unionSeq _ a b = instFunOf "union" (conjoinInst "and" a b);
    append _ a b = instInfix ["appended to the end of"] b a;
    prepend _ a b = instInfix ["appended to the start of"] b a; 
    subSequence _ a start end = np2inst ( -- here we have the problems of indices...
			postfixIndFun (["ranging from index"] ++ start.s!NomP ++ ["to index"] ++ end.s!NomP) 
			(appFunSg (funOfReg "subsequence") a)
	); 
    at _ a index = instFunOf (["element at index"] ++ (inst2np index).s!NomP) a; -- NomP?
    first _  = instFunOf ["first element"] ;
    last _ = instFunOf ["last element"];
    includingSeq _ = collIncluding "sequence";
    excludingSeq _ = collExcluding "sequence";
    selectSeq _ = collReject "sequence";
    rejectSeq _ = collReject "sequence";
    collectSeq _ _ = collCollect "sequence";
    seqAsBag _ = postfixIndFun ["as a bag"];
    seqAsSet _ = postfixIndFun ["as a set"];



