// $ANTLR 2.7.2: "decls.g" -> "DeclParser.java"$


    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.rule.*;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.logic.sort.*;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.java.visitor.*;
    import java.util.Vector;
    import java.util.HashMap;
    import de.uka.ilkd.key.proof.UseInsteadMap;


import antlr.TokenBuffer;
import antlr.TokenStreamException;
import antlr.TokenStreamIOException;
import antlr.ANTLRException;
import antlr.LLkParser;
import antlr.Token;
import antlr.TokenStream;
import antlr.RecognitionException;
import antlr.NoViableAltException;
import antlr.MismatchedTokenException;
import antlr.SemanticException;
import antlr.ParserSharedInputState;
import antlr.collections.impl.BitSet;

/** 
 * A parser for signature declarations,
 * i.e. sorts, schema variabls, predicates, functions and heuristics.
 */
public class DeclParser extends antlr.LLkParser       implements DeclParserTokenTypes
 {

	static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];

	private Namespace var_ns;
	private Namespace func_ns;
	private Namespace sort_ns;
	private Namespace heur_ns;
        private JavaInfo javaInfo;
	
	// these variables are set if a file is read in step by
	// step. This used when reading in LDTs because of cyclic
	// dependencies.
	protected boolean skip_schemavariables;
	protected boolean skip_functions;
	protected boolean skip_predicates;
	protected boolean skip_sorts;
	protected boolean skip_heuristics;
	protected boolean minus=false;	
	protected UseInsteadMap useInstead=new UseInsteadMap();


	public DeclParser(TokenStream lexer,
			  String filename,
			  Namespace var_ns,
			  Namespace func_ns,
			  Namespace sort_ns,
			  Namespace heur_ns,
                          JavaInfo ji) {
	    this(lexer);
	    setFilename(filename);
	    this.var_ns = var_ns;
	    this.func_ns = func_ns;
	    this.sort_ns = sort_ns;
	    this.heur_ns = heur_ns;
            this.javaInfo = ji;
	    setSkips(false, false, false, false, false);
	}
		
	public Namespace getSorts() {
	    return sort_ns;
	}

	public Namespace functions() {
	    return func_ns;
	}

	public Namespace heuristics() {
	    return heur_ns;
	}

	public Namespace variables() {
	    return var_ns;
	}


	protected int getLine() {
	    int line = -1;
	    try {
		line = LT(0).getLine();
	    } 
	    catch (TokenStreamException e) {
		System.err.println("No further token in stream");
	    }
	    return line;
	}   

	protected int getColumn() {
	    int col = -1;
	    try {
		col = LT(0).getColumn();
	    } 
	    catch (TokenStreamException e) {
		System.err.println("No further token in stream");
	    }
	    return col;
	}   


        protected void setSkips(boolean svars, boolean funcs, 
				boolean preds, boolean sorts,
				boolean heurs) {
	    skip_schemavariables = svars;
   	    skip_functions  	 = funcs;
	    skip_predicates 	 = preds;
	    skip_sorts 		 = sorts;
	    skip_heuristics 	 = heurs;
	}

	public void parseVariables() 	
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(false, true, true, true, true); 
		top();
		setSkips(false, false, false, false, false);	    		
	    }
    	public Term toZNotation(String s, boolean minus){
	Term[] t1=new Term[1];
	Operator v;
	Term a;
	TermFactory tf=TermFactory.DEFAULT;
	v=(Function) func_ns.lookup(new Name("#"));
	t1[0] = tf.createFunctionTerm((Function)v);
	v = (Function) func_ns.lookup(new Name(s.substring(0,1)));
	a = tf.createFunctionTerm((Function)v,t1);
	for(int i=1;i<s.length();i++){
		t1[0]=a;
		v = (Function)func_ns.lookup(new Name(s.substring(i,i+1)));
		a = tf.createFunctionTerm((Function)v,t1);
	}
	if(minus){
		t1[0]=a;
		v = (Function) func_ns.lookup(new Name("neg"));
		a = tf.createFunctionTerm((Function)v,t1);
	}
	t1[0]=a;
	v = (Function) func_ns.lookup(new Name("Z"));
	a = tf.createFunctionTerm((Function)v,t1);
	return a;
	}
	public void parseFunctions()
	    throws antlr.RecognitionException, antlr.TokenStreamException

	    {
		setSkips(true, false, true, true, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

	public void parsePredicates()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, false, true, true); 
		top();
		setSkips(false, false, false, false, false);	    
	    }

	public void parseSorts() 
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, true, false, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

	public void parseHeuristics()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, true, true, false); 
		top();
		setSkips(false, false, false, false, false);	    
	    }

	public void parseFuncAndPred()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, false, false, true, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

    public NamespaceSet namespaces() {
	NamespaceSet ns = new NamespaceSet();
	ns.setVariables(variables());
	ns.setFunctions(functions());
	ns.setSorts(getSorts());
	ns.setHeuristics(heuristics());
	return ns;
    }


protected DeclParser(TokenBuffer tokenBuf, int k) {
  super(tokenBuf,k);
  tokenNames = _tokenNames;
}

public DeclParser(TokenBuffer tokenBuf) {
  this(tokenBuf,3);
}

protected DeclParser(TokenStream lexer, int k) {
  super(lexer,k);
  tokenNames = _tokenNames;
}

public DeclParser(TokenStream lexer) {
  this(lexer,3);
}

public DeclParser(ParserSharedInputState state) {
  super(state,3);
  tokenNames = _tokenNames;
}

	public final void top() throws RecognitionException, TokenStreamException {
		
		
		decls();
	}
	
	public final void decls() throws RecognitionException, TokenStreamException {
		
		
		{
		_loop4:
		do {
			switch ( LA(1)) {
			case SORTS:
			{
				sort_decls();
				break;
			}
			case SCHEMA:
			{
				schema_var_decls();
				break;
			}
			case PROGRAM:
			{
				prog_var_decls();
				break;
			}
			case PREDICATES:
			{
				pred_decls();
				break;
			}
			case FUNCTIONS:
			{
				func_decls();
				break;
			}
			case HEURISTICS:
			{
				heur_decls();
				break;
			}
			default:
			{
				break _loop4;
			}
			}
		} while (true);
		}
	}
	
	public final void sort_decls() throws RecognitionException, TokenStreamException {
		
		
		boolean isObjectSort=false;
		boolean isGenericSort=false;
		Sort[] sortExt=new Sort [0];
		Sort[] sortOneOf=new Sort [0];
		ListOfName sort_ids;
		
		
		match(SORTS);
		match(LBRACE);
		{
		_loop10:
		do {
			if ((LA(1)==OBJECT||LA(1)==GENERIC||LA(1)==IDENT)) {
				if ( inputState.guessing==0 ) {
					
						       isObjectSort=false; 
						       isGenericSort=false; 
						       sortExt = new Sort[0];
						       sortOneOf = new Sort[0];
						
				}
				{
				switch ( LA(1)) {
				case OBJECT:
				{
					match(OBJECT);
					if ( inputState.guessing==0 ) {
						isObjectSort=true;
					}
					sort_ids=sort_identifiers();
					break;
				}
				case GENERIC:
				{
					match(GENERIC);
					if ( inputState.guessing==0 ) {
						isGenericSort=true;
					}
					sort_ids=sort_identifiers();
					{
					switch ( LA(1)) {
					case ONEOF:
					{
						match(ONEOF);
						sortOneOf=oneof_sorts();
						break;
					}
					case EXTENDS:
					case SEMI:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					{
					switch ( LA(1)) {
					case EXTENDS:
					{
						match(EXTENDS);
						sortExt=extends_sorts();
						break;
					}
					case SEMI:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					break;
				}
				case IDENT:
				{
					sort_ids=sort_identifiers();
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				match(SEMI);
				if ( inputState.guessing==0 ) {
					
						    Name           sort_name;
						    IteratorOfName it         = sort_ids.iterator ();
						    while ( it.hasNext () ) {
						    sort_name = it.next ();
					
						    if (sort_ns.lookup(sort_name) == null) {
							Sort s;
							if (isObjectSort) {
							    if ("Object".equals(sort_name)) {
								if (sort_ns.lookup(new Name("Object"))
								    ==null) {
					
								    s = new ClassInstanceSortImpl(sort_name);
								}
								s=(Sort)sort_ns.lookup(new Name("Object"));
							    } else {
								s = new ClassInstanceSortImpl(sort_name, 
											      (Sort)sort_ns.lookup
											      (new Name("Object")));
							    }	
							} else if (isGenericSort) {
							    int        i;
							    SetOfSort  ext   = GenericSort.EMPTY_SORT_SET;
							    SetOfSort  oneOf = GenericSort.EMPTY_SORT_SET;
					
							    for ( i = 0; i != sortExt.length; ++i )
							        ext = ext.add ( sortExt[i] );
					
							    for ( i = 0; i != sortOneOf.length; ++i )
							        oneOf = oneOf.add ( sortOneOf[i] );
					
							    try {
							        s = new GenericSort ( sort_name,
							                              ext,
							                              oneOf );
							    } catch ( GenericSupersortException e ) {
							        throw new GenericSortException ( "sort",
							                   "Illegal sort given", e.getIllegalSort (),
							                   getFilename (), getLine (), getColumn () );
							    }
							} else {
							    s = new PrimitiveSort(sort_name);
							}
							sort_ns.add(s);
							if (s instanceof NonCollectionSort) {
							    NonCollectionSort ns
								= (NonCollectionSort)s;
							    sort_ns.add(ns.getSetSort());
							    sort_ns.add(ns.getSequenceSort());
							    sort_ns.add(ns.getBagSort());
							}
							if ( s instanceof SortDefiningSymbols )
							    ((SortDefiningSymbols)s).createSymbols ( func_ns, sort_ns );
						    }
					
						    }
						
				}
			}
			else {
				break _loop10;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void schema_var_decls() throws RecognitionException, TokenStreamException {
		
		
			Sort s = null;
			boolean makeVariableSV  = false;
			boolean makeDependingSV = false; // skolem functions
			boolean rigid           = false;
			boolean listModifier    = false;
		
		
		match(SCHEMA);
		match(VARIABLES);
		match(LBRACE);
		{
		_loop39:
		do {
			if ((LA(1)==PROGRAM||LA(1)==FORMULA||LA(1)==IDENT)) {
				if ( inputState.guessing==0 ) {
					makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;
				}
				{
				switch ( LA(1)) {
				case PROGRAM:
				{
					{
					match(PROGRAM);
					{
					switch ( LA(1)) {
					case SVLIST:
					{
						match(SVLIST);
						if ( inputState.guessing==0 ) {
							listModifier=true;
						}
						break;
					}
					case ARRAY:
					case STATEMENT:
					case EXPRESSION:
					case STATEMENTEXPRESSION:
					case BRANCH:
					case NONSIMPLE:
					case LEFTHANDSIDE:
					case SIMPLE:
					case LOOPINIT:
					case GUARD:
					case FORUPDATES:
					case TYPE:
					case JAVABOOL:
					case VARIABLE:
					case SWITCH:
					case BOOLEANVARIABLE:
					case CATCH:
					case METHODREFERENCE:
					case SPECIAL:
					case CLASSTYPE:
					case INTERFACETYPE:
					case ARRAYTYPE:
					case METHODBODY:
					case VARIABLEINIT:
					case LABEL:
					case MULTIPLEVARDECL:
					case NEW:
					case IMPLICIT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					{
					switch ( LA(1)) {
					case STATEMENT:
					{
						match(STATEMENT);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.STATEMENT;
												
						}
						break;
					}
					case EXPRESSION:
					{
						match(EXPRESSION);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.EXPRESSION;
												
						}
						break;
					}
					case MULTIPLEVARDECL:
					{
						match(MULTIPLEVARDECL);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.MULTIPLEVARDECL;
												
						}
						break;
					}
					case JAVABOOL:
					{
						match(JAVABOOL);
						match(EXPRESSION);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.BOOLEANEXPRESSION;
												
						}
						break;
					}
					case LEFTHANDSIDE:
					{
						match(LEFTHANDSIDE);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.LEFTHANDSIDE;
												
						}
						break;
					}
					case LOOPINIT:
					{
						match(LOOPINIT);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.LOOPINIT;
												
						}
						break;
					}
					case GUARD:
					{
						match(GUARD);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.GUARD;
												
						}
						break;
					}
					case FORUPDATES:
					{
						match(FORUPDATES);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.FORUPDATES;
												
						}
						break;
					}
					case STATEMENTEXPRESSION:
					{
						match(STATEMENTEXPRESSION);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.STATEMENTEXPRESSION;
												
						}
						break;
					}
					case LABEL:
					{
						match(LABEL);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.LABEL;
												
						}
						break;
					}
					case TYPE:
					{
						match(TYPE);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.TYPE;
												
						}
						break;
					}
					case VARIABLE:
					{
						match(VARIABLE);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.VARIABLE;
												
						}
						break;
					}
					case VARIABLEINIT:
					{
						match(VARIABLEINIT);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.VARIABLEINIT;
												
						}
						break;
					}
					case SWITCH:
					{
						match(SWITCH);
						if ( inputState.guessing==0 ) {
							
									 	s=ProgramSVSort.SWITCH;
										
						}
						break;
					}
					case BOOLEANVARIABLE:
					{
						{
						match(BOOLEANVARIABLE);
						}
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.BOOLEANVARIABLE;
												
						}
						break;
					}
					case CATCH:
					{
						{
						match(CATCH);
						}
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.CATCH;
												
						}
						break;
					}
					case METHODREFERENCE:
					{
						{
						match(METHODREFERENCE);
						}
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.METHODREFERENCE;
												
						}
						break;
					}
					case SPECIAL:
					{
						match(SPECIAL);
						match(CONSTRUCTORREFERENCE);
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.SPECIALCONSTRUCTORREFERENCE;				
												
						}
						break;
					}
					case ARRAYTYPE:
					{
						{
						match(ARRAYTYPE);
						}
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.ARRAYTYPE;
												
						}
						break;
					}
					case METHODBODY:
					{
						{
						match(METHODBODY);
						}
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.METHODBODY;
												
						}
						break;
					}
					case BRANCH:
					{
						{
						match(BRANCH);
						}
						if ( inputState.guessing==0 ) {
							
													s=ProgramSVSort.BRANCH;
												
						}
						break;
					}
					default:
						if ((LA(1)==SIMPLE) && (LA(2)==EXPRESSION)) {
							match(SIMPLE);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVABOOL)) {
							match(SIMPLE);
							match(JAVABOOL);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVABOOLEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVABYTE)) {
							match(SIMPLE);
							match(JAVABYTE);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVABYTEEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVACHAR)) {
							match(SIMPLE);
							match(JAVACHAR);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVACHAREXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVASHORT)) {
							match(SIMPLE);
							match(JAVASHORT);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVASHORTEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVAINT)) {
							match(SIMPLE);
							match(JAVAINT);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVAINTEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVALONG)) {
							match(SIMPLE);
							match(JAVALONG);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVALONGEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==JAVABYTESHORTINT)) {
							match(SIMPLE);
							match(JAVABYTESHORTINT);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEJAVABYTESHORTINTEXPRESSION;
													
							}
						}
						else if ((LA(1)==SIMPLE) && (LA(2)==ANYJAVATYPE)) {
							match(SIMPLE);
							match(ANYJAVATYPE);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.SIMPLEANYJAVATYPEEXPRESSION;
													
							}
						}
						else if ((LA(1)==CLASSTYPE) && (LA(2)==EXPRESSION)) {
							match(CLASSTYPE);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.CLASSTYPEEXPRESSION;
													
							}
						}
						else if ((LA(1)==INTERFACETYPE) && (LA(2)==EXPRESSION)) {
							match(INTERFACETYPE);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.INTERFACETYPEEXPRESSION;
													
							}
						}
						else if ((LA(1)==NONSIMPLE) && (LA(2)==EXPRESSION)) {
							match(NONSIMPLE);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.NONSIMPLEEXPRESSION;
													
							}
						}
						else if ((LA(1)==NONSIMPLE) && (LA(2)==JAVABOOL)) {
							match(NONSIMPLE);
							match(JAVABOOL);
							match(EXPRESSION);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.NONSIMPLEBOOLEANEXPRESSION;
													
							}
						}
						else if ((LA(1)==IMPLICIT) && (LA(2)==NEXT)) {
							match(IMPLICIT);
							match(NEXT);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.IMPLICITNEXT;
													
							}
						}
						else if ((LA(1)==IMPLICIT) && (LA(2)==REFERENCE)) {
							match(IMPLICIT);
							match(REFERENCE);
							match(FIELD);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.IMPLICITREFERENCE;
													
							}
						}
						else if ((LA(1)==NEW) && (LA(2)==IDENT)) {
							match(NEW);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.NEW;
													
							}
						}
						else if ((LA(1)==NEW) && (LA(2)==ARRAY)) {
							match(NEW);
							match(ARRAY);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.NEWARRAY;
													
							}
						}
						else if ((LA(1)==ARRAY) && (LA(2)==INITIALIZER)) {
							match(ARRAY);
							match(INITIALIZER);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.ARRAYINITIALIZER;
													
							}
						}
						else if ((LA(1)==ARRAY) && (LA(2)==LENGTH)) {
							match(ARRAY);
							match(LENGTH);
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.ARRAYLENGTH;
													
							}
						}
						else if ((LA(1)==CLASSTYPE) && (LA(2)==IDENT)) {
							{
							match(CLASSTYPE);
							}
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.CLASSTYPE;
													
							}
						}
						else if ((LA(1)==INTERFACETYPE) && (LA(2)==IDENT)) {
							{
							match(INTERFACETYPE);
							}
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.INTERFACETYPE;
													
							}
						}
						else if ((LA(1)==ARRAY) && (LA(2)==IDENT)) {
							{
							match(ARRAY);
							}
							if ( inputState.guessing==0 ) {
								
														s=ProgramSVSort.ARRAY;
													
							}
						}
					else {
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					}
					break;
				}
				case FORMULA:
				{
					match(FORMULA);
					{
					switch ( LA(1)) {
					case RIGID:
					{
						match(RIGID);
						if ( inputState.guessing==0 ) {
							rigid = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					if ( inputState.guessing==0 ) {
						
											s = Sort.FORMULA;
										
					}
					break;
				}
				case IDENT:
				{
					s=any_sort_id_check(true);
					{
					switch ( LA(1)) {
					case VARIABLES:
					{
						match(VARIABLES);
						if ( inputState.guessing==0 ) {
							makeVariableSV = true;
						}
						break;
					}
					case DEPENDING:
					{
						match(DEPENDING);
						if ( inputState.guessing==0 ) {
							makeDependingSV = true;
						}
						break;
					}
					case RIGID:
					{
						match(RIGID);
						if ( inputState.guessing==0 ) {
							rigid = true;
						}
						break;
					}
					case IDENT:
					{
						break;
					}
					default:
					{
						throw new NoViableAltException(LT(1), getFilename());
					}
					}
					}
					break;
				}
				default:
				{
					throw new NoViableAltException(LT(1), getFilename());
				}
				}
				}
				one_schema_var_decl(s,makeVariableSV,makeDependingSV,rigid,listModifier);
				{
				_loop38:
				do {
					if ((LA(1)==COMMA)) {
						match(COMMA);
						one_schema_var_decl(s,makeVariableSV,makeDependingSV,rigid,listModifier);
					}
					else {
						break _loop38;
					}
					
				} while (true);
				}
				match(SEMI);
			}
			else {
				break _loop39;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void prog_var_decls() throws RecognitionException, TokenStreamException {
		
		
			Sort s;
		String var_name;
		
		
		match(PROGRAM);
		match(VARIABLES);
		match(LBRACE);
		{
		_loop53:
		do {
			if ((LA(1)==IDENT)) {
				s=sort_id_check(true);
				var_name=id();
				if ( inputState.guessing==0 ) {
					
									ProgramVariable pv = new ProgramVariable(
					new ProgramElementName(var_name), 
					javaInfo.getKeYJavaType(s));
									javaInfo.programVariables().add(pv);
								
				}
				match(SEMI);
			}
			else {
				break _loop53;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void pred_decls() throws RecognitionException, TokenStreamException {
		
		
			Sort[] argSorts;
			String pred_name;
		
		
		match(PREDICATES);
		match(LBRACE);
		{
		_loop43:
		do {
			if ((LA(1)==IDENT)) {
				pred_name=id();
				argSorts=arg_sorts(!skip_predicates);
				if ( inputState.guessing==0 ) {
					
								    if (!skip_predicates) {
									Name predicate = new Name(pred_name);
									if (func_ns.lookup(predicate) != null) {
									    // throw new AmbigiousDeclException
									    //("predicate "+pred_name+" ", 
									    //getLine(), getColumn());
									} else {
									    Function p = new Function(predicate,
												      Sort.FORMULA, 
												      argSorts);
									    func_ns.add(p);			
									}
								    } 
								
				}
				match(SEMI);
			}
			else {
				break _loop43;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void func_decls() throws RecognitionException, TokenStreamException {
		
		
			Sort[] argSorts;
			Sort retSort;
		String func_name;
		
		
		match(FUNCTIONS);
		match(LBRACE);
		{
		_loop46:
		do {
			if ((LA(1)==IDENT)) {
				retSort=sort_id_check(!skip_functions);
				func_name=id();
				argSorts=arg_sorts(!skip_functions);
				if ( inputState.guessing==0 ) {
					
								    if (!skip_functions) {
					
									Name fct_name = new Name(func_name);
									if (func_ns.lookup(fct_name) != null) {
									    //throw new AmbigiousDeclException
									    //("function "+func_name, 
									    //getLine(), getColumn());
									} else {
									    Function f=new Function(fct_name,
												    retSort, 
												    argSorts);
									    func_ns.add(f);
									}
								    } 
								
				}
				match(SEMI);
			}
			else {
				break _loop46;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final void heur_decls() throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		match(HEURISTICS);
		match(LBRACE);
		{
		_loop56:
		do {
			if ((LA(1)==IDENT)) {
				id = LT(1);
				match(IDENT);
				match(SEMI);
				if ( inputState.guessing==0 ) {
					
								    if (!skip_heuristics) {
									Heuristic h = new Heuristic(new Name(id.getText()));
									heur_ns.add(h);
								    }
								
				}
			}
			else {
				break _loop56;
			}
			
		} while (true);
		}
		match(RBRACE);
	}
	
	public final ListOfName  sort_identifiers() throws RecognitionException, TokenStreamException {
		ListOfName ids = SLListOfName.EMPTY_LIST;
		
		Token  id = null;
		Token  id2 = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			ids = ids.append ( new Name(id.getText()) );
		}
		{
		_loop13:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				id2 = LT(1);
				match(IDENT);
				if ( inputState.guessing==0 ) {
					ids = ids.append ( new Name(id2.getText()) );
				}
			}
			else {
				break _loop13;
			}
			
		} while (true);
		}
		return ids;
	}
	
	public final Sort[]  oneof_sorts() throws RecognitionException, TokenStreamException {
		Sort[] oneOfSorts = null;
		
		
			Vector args = new Vector();
			Sort s;
		
		
		match(LBRACE);
		s=sort_id_check(true);
		if ( inputState.guessing==0 ) {
			args.add(s);
		}
		{
		_loop19:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s=sort_id_check(true);
				if ( inputState.guessing==0 ) {
					args.add(s);
				}
			}
			else {
				break _loop19;
			}
			
		} while (true);
		}
		match(RBRACE);
		if ( inputState.guessing==0 ) {
			
						oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
					
		}
		return oneOfSorts;
	}
	
	public final Sort[]  extends_sorts() throws RecognitionException, TokenStreamException {
		Sort[] extendsSorts = null;
		
		
			Vector args = new Vector();
			Sort s;
		
		
		s=any_sort_id_check(true);
		if ( inputState.guessing==0 ) {
			args.add(s);
		}
		{
		_loop16:
		do {
			if ((LA(1)==COMMA)) {
				match(COMMA);
				s=any_sort_id_check(true);
				if ( inputState.guessing==0 ) {
					args.add(s);
				}
			}
			else {
				break _loop16;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			
						extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
					
		}
		return extendsSorts;
	}
	
	public final Sort  any_sort_id_check(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		Sort t;
		
		
		t=any_sort_id_check_help(checkSort);
		s=array_decls(t);
		return s;
	}
	
	public final Sort  sort_id_check(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		Sort t;
		
		
		t=sort_id_check_help(checkSort);
		s=array_decls(t);
		return s;
	}
	
	public final void one_schema_var_decl(
		Sort s, 
	            boolean makeVariableSV, 
	            boolean makeDependingSV,
	            boolean rigid, 
	            boolean listModifier
	) throws RecognitionException, TokenStreamException {
		
		Token  id = null;
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
			
					    if (!skip_schemavariables) {
						SchemaVariable v;
						if ( s == Sort.FORMULA ) {
						    v = SchemaVariable.createFormulaSV
							(new Name(id.getText()), listModifier, rigid);
						} else if ( s instanceof ProgramSVSort ) {
						    v = SchemaVariable.createProgramSV
						    (new ProgramElementName(id.getText()),(ProgramSVSort) s, listModifier);
						} else {
						    if ( makeVariableSV ) {
							v=SchemaVariable.createVariableSV
							(new Name(id.getText()), s, listModifier);
						    } else if ( makeDependingSV ) {
							v = SchemaVariable.createDependingSV
							(new Name(id.getText()), s, listModifier);
						    } else { v = SchemaVariable.createTermSV
							     (new Name(id.getText()), s, listModifier, rigid);
						    }
						}
						var_ns.add(v);
					    }
					
		}
	}
	
	public final String  id() throws RecognitionException, TokenStreamException {
		String name = null;
		
		Token  classID = null;
		Token  nameID = null;
		Token  pureID = null;
		
		boolean synPredMatched69 = false;
		if (((LA(1)==IDENT) && (LA(2)==COLON))) {
			int _m69 = mark();
			synPredMatched69 = true;
			inputState.guessing++;
			try {
				{
				match(IDENT);
				match(COLON);
				match(COLON);
				match(IDENT);
				}
			}
			catch (RecognitionException pe) {
				synPredMatched69 = false;
			}
			rewind(_m69);
			inputState.guessing--;
		}
		if ( synPredMatched69 ) {
			classID = LT(1);
			match(IDENT);
			match(COLON);
			match(COLON);
			nameID = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				
							 name = classID.getText() + "::" + nameID.getText();
							
			}
		}
		else if ((LA(1)==IDENT) && (LA(2)==SEMI||LA(2)==LPAREN)) {
			pureID = LT(1);
			match(IDENT);
			if ( inputState.guessing==0 ) {
				name = pureID.getText();
			}
		}
		else {
			throw new NoViableAltException(LT(1), getFilename());
		}
		
		return name;
	}
	
	public final Sort[]  arg_sorts(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort[] argSorts = null;
		
		
			Vector args = new Vector();
			Sort s;
		
		
		{
		switch ( LA(1)) {
		case LPAREN:
		{
			match(LPAREN);
			s=sort_id_check(checkSort);
			if ( inputState.guessing==0 ) {
				args.add(s);
			}
			{
			_loop50:
			do {
				if ((LA(1)==COMMA)) {
					match(COMMA);
					s=sort_id_check(checkSort);
					if ( inputState.guessing==0 ) {
						args.add(s);
					}
				}
				else {
					break _loop50;
				}
				
			} while (true);
			}
			match(RPAREN);
			break;
		}
		case SEMI:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
						argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
					
		}
		return argSorts;
	}
	
	public final Sort  sort_id() throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		s=sort_id_check(true);
		return s;
	}
	
	public final Sort  sort_id_check_help(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		s=any_sort_id_check_help(checkSort);
		if ( inputState.guessing==0 ) {
			
			// don't allow generic sorts or collection sorts of
			// generic sorts at this point
			Sort t = s;
			while ( t instanceof CollectionSort )
			t = ((CollectionSort)t).elementSort ();
			if ( t instanceof GenericSort )
			throw new GenericSortException ( "sort",
			"Non-generic sort expected", t,
			getFilename (), getLine (), getColumn () );
			
		}
		return s;
	}
	
	public final Sort  array_decls(
		Sort p
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		
		int  n=0;
		
		
		{
		_loop64:
		do {
			if ((LA(1)==LBRACKET)) {
				match(LBRACKET);
				match(RBRACKET);
				if ( inputState.guessing==0 ) {
					n++;
				}
			}
			else {
				break _loop64;
			}
			
		} while (true);
		}
		if ( inputState.guessing==0 ) {
			s = p;
			while(n>0){
			s=ArraySortImpl.getArraySort(s);
			n--;
			}
					
		}
		return s;
	}
	
	public final Sort  any_sort_id_check_help(
		boolean checkSort
	) throws RecognitionException, TokenStreamException {
		Sort s = null;
		
		Token  id = null;
		
		String name;
		
		
		id = LT(1);
		match(IDENT);
		if ( inputState.guessing==0 ) {
				String idtext=(useInstead.get(id.getText()));
						s= (Sort) sort_ns.lookup(new Name(idtext));
						if (checkSort && s == null) {
						    throw 
							new NotDeclException("sort", id, 
									     getFilename());
						}
					
		}
		return s;
	}
	
	public final IdDeclaration  id_declaration() throws RecognitionException, TokenStreamException {
		 IdDeclaration idd = null ;
		
		Token  id = null;
		
		Sort s = null;
		
		
		id = LT(1);
		match(IDENT);
		{
		switch ( LA(1)) {
		case COLON:
		{
			match(COLON);
			s=sort_id_check(true);
			break;
		}
		case EOF:
		{
			break;
		}
		default:
		{
			throw new NoViableAltException(LT(1), getFilename());
		}
		}
		}
		if ( inputState.guessing==0 ) {
			
			idd = new IdDeclaration ( id.getText (), s );
			
		}
		return idd;
	}
	
	
	public static final String[] _tokenNames = {
		"<0>",
		"EOF",
		"<2>",
		"NULL_TREE_LOOKAHEAD",
		"\"java\"",
		"\"array\"",
		"\"initializer\"",
		"\"statement\"",
		"\"expression\"",
		"\"statementexpression\"",
		"\"branch\"",
		"\"nonsimple\"",
		"\"lefthandside\"",
		"\"simple\"",
		"\"notsimple\"",
		"\"loopinit\"",
		"\"guard\"",
		"\"forupdates\"",
		"\"typeof\"",
		"\"type\"",
		"\"javaboolean\"",
		"\"object\"",
		"\"javabyte\"",
		"\"javashort\"",
		"\"javaint\"",
		"\"javalong\"",
		"\"javachar\"",
		"\"javabyteshortint\"",
		"\"anyjavatype\"",
		"\"number\"",
		"\"intliteral\"",
		"\"longliteral\"",
		"\"generic\"",
		"\"extends\"",
		"\"oneof\"",
		"\"variable\"",
		"\"switch\"",
		"\"booleanvariable\"",
		"\"catch\"",
		"\"methodreference\"",
		"\"constructorreference\"",
		"\"special\"",
		"\"classtype\"",
		"\"interfacetype\"",
		"\"arraytype\"",
		"\"methodbody\"",
		"\"variableinitializer\"",
		"\"label\"",
		"\"program\"",
		"\"multiplevardecl\"",
		"\"svlist\"",
		"\"all\"",
		"\"ex\"",
		"\"true\"",
		"\"false\"",
		"\"find\"",
		"\"if\"",
		"\"varcond\"",
		"\"noninteractive\"",
		"\"recursive\"",
		"\"displayname\"",
		"\"helptext\"",
		"\"new\"",
		"\"not\"",
		"\"free\"",
		"\"in\"",
		"\"depending\"",
		"\"on\"",
		"\"close\"",
		"\"goal\"",
		"\"replacewith\"",
		"\"add\"",
		"\"addrules\"",
		"\"addprogvars\"",
		"\"heuristics\"",
		"\"sorts\"",
		"\"schema\"",
		"\"variables\"",
		"\"implicit\"",
		"\"field\"",
		"\"reference\"",
		"\"next\"",
		"\"length\"",
		"\"rigid\"",
		"\"predicates\"",
		"\"functions\"",
		"\"formula\"",
		"\"rules\"",
		"\"proof\"",
		"\"rule\"",
		"\"term\"",
		"\"inst\"",
		"\"ifseqformula\"",
		"\"interactive\"",
		"\"heur\"",
		"\"problem\"",
		"\"#inType\"",
		"\"#isObject\"",
		"\"#isClonable\"",
		"\"#inLong\"",
		"\"#inInt\"",
		"\"#inChar\"",
		"\"#inShort\"",
		"\"#inByte\"",
		"\"#add\"",
		"\"#sub\"",
		"\"#mul\"",
		"\"#div\"",
		"\"#mod\"",
		"\"#less\"",
		"\"#greater\"",
		"\"#leq\"",
		"\"#geq\"",
		"\"#eq\"",
		"\"#JavaIntUnaryMinus\"",
		"\"#JavaIntAdd\"",
		"\"#JavaIntSub\"",
		"\"#JavaIntMul\"",
		"\"#JavaIntDiv\"",
		"\"#JavaIntMod\"",
		"\"#JavaIntAnd\"",
		"\"#JavaIntOr\"",
		"\"#JavaIntXor\"",
		"\"#JavaIntComplement\"",
		"\"#JavaIntShiftRight\"",
		"\"#JavaIntShiftLeft\"",
		"\"#JavaIntUnsignedShiftRight\"",
		"\"#JavaLongUnaryMinus\"",
		"\"#JavaLongAdd\"",
		"\"#JavaLongSub\"",
		"\"#JavaLongMul\"",
		"\"#JavaLongDiv\"",
		"\"#JavaLongMod\"",
		"\"#JavaLongAnd\"",
		"\"#JavaLongOr\"",
		"\"#JavaLongXor\"",
		"\"#JavaLongComplement\"",
		"\"#JavaLongShiftRight\"",
		"\"#JavaLongShiftLeft\"",
		"\"#JavaLongUnsignedShiftRight\"",
		"\"#moduloByte\"",
		"\"#moduloShort\"",
		"\"#moduloInteger\"",
		"\"#moduloLong\"",
		"\"#staticanalysis\"",
		"\"#lengthReference\"",
		"\"waryAll\"",
		"\"waryEx\"",
		"\"sameUpdateLevel\"",
		"VOCAB",
		"`;'",
		"`/'",
		"`:'",
		"`/->'",
		"`:='",
		"`.'",
		"`,'",
		"`('",
		"`)'",
		"`{'",
		"`}'",
		"`['",
		"']'",
		"`at'",
		"`|'",
		"`&'",
		"`->'",
		"`<->'",
		"`='",
		"`==>'",
		"'^'",
		"java operator",
		"white space",
		"a string in double quotes",
		"a char in single quotes",
		"ESC",
		"DIAMOND",
		"BOX",
		"BLOCK",
		"a string with double quotes",
		"a comment",
		"a comment",
		"a digit",
		"a hexadeciamal number",
		"a letter",
		"an admissible character for identifiers",
		"an identifer"
	};
	
	
	}
