header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.parser.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;    
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.rule.*;

    import de.uka.ilkd.key.java.SchemaJavaReader;
    import java.util.Vector;
    import java.util.HashMap;
    import java.util.HashSet;
    import java.util.Iterator;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.proof.UseInsteadMap;
    import de.uka.ilkd.key.java.Services;
    import de.uka.ilkd.key.gui.AbbrevMap;
}
class TacletParser extends Parser;

options {
	importVocab=KeYLexer;
	k= 2;
	defaultErrorHandler=false;
}

{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];
    protected boolean minus=false;
    protected Namespace var_ns;
    protected Namespace func_ns;
    protected Namespace sort_ns;
    protected Namespace heur_ns;
    protected JavaInfo javaInfo;  
    protected AbbrevMap scm = new AbbrevMap();

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_heuristics;
    protected UseInsteadMap useInstead=new UseInsteadMap();

    protected TermFactory tf;
    protected SchemaJavaReader javaReader;
    boolean schematicJava;	

    public boolean errors = false;

    public TacletParser(TokenStream lexer,
		      String filename,
		      TermFactory tf,
		      SchemaJavaReader jr,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer,2);
	setFilename(filename);
	this.tf = tf;
	this.javaReader = jr;
	this.var_ns  = var_ns;
	this.func_ns = func_ns;
	this.sort_ns = sort_ns;
	this.heur_ns = heur_ns;
	schematicJava=true;	
    }

    public TacletParser(TokenStream lexer,
		      String filename,
		      SchemaJavaReader jr,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer, filename, TermFactory.DEFAULT, 
	     jr, var_ns, func_ns, sort_ns, heur_ns);
    }


    public TacletParser(TokenStream lexer,
		      String filename,
		      TermFactory tf,
		      JavaInfo javaInfo,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer, filename, tf, 
	     new de.uka.ilkd.key.java.SchemaRecoder2KeY(javaInfo.rec2key(),
						javaInfo.getTypeConverter(),
						new NamespaceSet(var_ns, func_ns, 
							sort_ns, heur_ns)), 	     
	     var_ns, func_ns, sort_ns,heur_ns);
    }

    public TacletParser(TokenStream lexer,
		      String filename,
	              JavaInfo javaInfo,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer, filename,TermFactory.DEFAULT,
	     new de.uka.ilkd.key.java.SchemaRecoder2KeY(javaInfo.rec2key(),
						javaInfo.getTypeConverter(),
						new NamespaceSet(var_ns, func_ns, 
							sort_ns, heur_ns)),
	     var_ns, func_ns, sort_ns, heur_ns);
    }


    private TacletBuilder createTacletBuilderFor
                (Object find, boolean sameUpdatePrefix) 
	throws InvalidFindException {
        if ( sameUpdatePrefix && !( find instanceof Term ) )
		throw new InvalidFindException
		    ("\"sameUpdateLevel\" may only be used for rewrite taclets:"+find,
		     getFilename(), getLine(), getColumn());
	if ( find == null ) {
	    return new NoFindTacletBuilder();
        } else if ( find instanceof Term ) {
	    return new RewriteTacletBuilder().setFind((Term)find)
	                     .setSameUpdatePrefix(sameUpdatePrefix);
        } else if ( find instanceof Sequent ) {
	    Sequent findSeq = (Sequent) find;
	    if ( findSeq == Sequent.EMPTY_SEQUENT ) {
		return new NoFindTacletBuilder();
	    } else if (   findSeq.antecedent().size() == 1
			  && findSeq.succedent().size() == 0 ) {
		Term findFma = findSeq.antecedent().get(0).formula();
		return new AntecTacletBuilder().setFind(findFma);
	    } else if (   findSeq.antecedent().size() == 0
			  && findSeq.succedent().size() == 1 ) {
		Term findFma = findSeq.succedent().get(0).formula();
		return new SuccTacletBuilder().setFind(findFma);
	    } else {
		throw new InvalidFindException
		    ("Unknown find-sequent (perhaps null?):"+findSeq,
		     getFilename(), getLine(), getColumn());
	    }
        } else {
	    throw new RuntimeException("createTacletBuilderFor called with"
				       + find.getClass().getName() + "object");
        }
    }	    

	protected static HashSet progVars(JavaBlock jb) {
		ProgramVariableCollector pvc
			= new ProgramVariableCollector(jb.program());
		pvc.start();
		return pvc.result();
	}


	protected void bindProgVars(HashSet progVars) {
		var_ns=new Namespace(var_ns);
		Iterator it=progVars.iterator();
		while (it.hasNext()) {
			var_ns.add((ProgramVariable)it.next());
		}
	}

	protected void unbindProgVars() {
		var_ns = var_ns.parent();
	}

    protected boolean isNumber(String s){
	      boolean num=true;
          if(s==null){
            num=false;
          }else{
	        for(int i=s.length()-1;i>=0;i--){
		     if(s.charAt(i)<'0'||s.charAt(i)>'9')
			    num=false;
	        }
         }
	     return num;
    }

    
    public Term toZNotation(String s, boolean minus){
	Term[] t1=new Term[1];
	Operator v;
	Term a;
	v=(Function) func_ns.lookup(new Name("#"));
	t1[0] = tf.createFunctionTerm((Function)v);
	v = (Function) func_ns.lookup(new Name(s.substring(0,1)));
	a = tf.createFunctionTerm((Function)v,t1);
	for(int i=1;i<s.length();i++){
		t1[0]=a;
		v = (Function)func_ns.lookup(new Name(s.substring(i,i+1)));
		a = tf.createFunctionTerm((Function)v,t1);
	}
	if(minus){
		t1[0]=a;
		v = (Function) func_ns.lookup(new Name("neg"));
		a = tf.createFunctionTerm((Function)v,t1);
	}
	t1[0]=a;
	v = (Function) func_ns.lookup(new Name("Z"));
	a = tf.createFunctionTerm((Function)v,t1);
	return a;
    }



    private static ListOfRulevariable NO_RULE_VARIABLES =
	SLListOfRulevariable.EMPTY_LIST;

    private void addGoalTemplate(TacletBuilder b,
                                 String id,
				 Object rwObj,
				 Sequent addSeq,
				 ListOfTaclet addRList,
				 SetOfSchemaVariable pvs) 
	throws SemanticException
	{
	    TacletGoalTemplate gt = null;
	    if ( rwObj == null ) {
		// there is no replacewith, so we take
		gt = new TacletGoalTemplate(addSeq,
					  addRList,
					  NO_RULE_VARIABLES,
					  pvs);
	    } else {
		if ( b instanceof NoFindTacletBuilder ) {
		    // there is a replacewith without a find.
		    throw 
			new UnfittingReplacewithException
			("Replacewith without find", getFilename(),
			 getLine(), getColumn());
		} else if ( b instanceof SuccTacletBuilder
			    || b instanceof AntecTacletBuilder ) {
		    if ( rwObj instanceof Sequent ) {
			gt = new AntecSuccTacletGoalTemplate(addSeq,
							   addRList,
							   NO_RULE_VARIABLES,
							   (Sequent)rwObj,
							   pvs);  
		    } else {
			throw new UnfittingReplacewithException
			    ("Replacewith in a Antec-or SuccTaclet has "+
			     "to contain a sequent (not a term)", 
			     getFilename(), getLine(), getColumn());
		    }
		} else if ( b instanceof RewriteTacletBuilder ) {
		    if ( rwObj instanceof Term ) {
			gt = new RewriteTacletGoalTemplate(addSeq,
							 addRList,
							 NO_RULE_VARIABLES,
							 (Term)rwObj,
							 pvs);  
		    } else {
			throw new UnfittingReplacewithException
			    ("Replacewith in a RewriteTaclet has "+
			     "to contain a term (not a sequent)", 
			     getFilename(), getLine(), getColumn());
		    }
		}
	    }
            gt.setName(id); 
	    b.addTacletGoalTemplate(gt);
	}
						


    protected int getLine() {
	int line = -1;
	try {
	    line = LT(0).getLine();
	} 
	catch (TokenStreamException e) {
	    System.err.println("No further token in stream");
	}
	return line;
    }   

    protected int getColumn() {
	int col = -1;
	try {
	    col = LT(0).getColumn();
	} 
	catch (TokenStreamException e) {
	    System.err.println("No further token in stream");
	}
	return col;
    }   


    protected Term termForParsedVariable(ParsableVariable v) 
	throws antlr.SemanticException 
	{
		if ( v instanceof LogicVariable ) {
			return tf.createVariableTerm((LogicVariable)v);
		} else if ( v instanceof SchemaVariable ) {
			return tf.createVariableTerm((SchemaVariable)v);
		} else if ( v instanceof ProgramVariable ) {
			return tf.createVariableTerm((ProgramVariable)v);
//		} else if ( v instanceof ParsableVariable ) {
//			return tf.createVariableTerm((ParsableVariable)v);

		} else {
			throw new antlr.SemanticException
			    (""+ v +" is not a logic or schema or program variable");
		}
	}

    protected TermSymbol progVar(Token id,Sort s) throws NotDeclException{
	TermSymbol v = (TermSymbol)
	    var_ns.lookup(new Name(id.getText()));

	if (v == null) {
		throw new NotDeclException("program variable", id,
						getFilename());
	}
	return v;
    }

    protected LogicVariable bindVar(Token id,Sort s) {
	LogicVariable v=new LogicVariable(new Name(id.getText()), s);
	var_ns=var_ns.extended(v);
	return v;
    }
	
    protected void unbindVars() {
	var_ns = var_ns.parent();
    }

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
	String s=t.getText();
	if (!schematicJava) {
		try {
			return javaReader.readBlockWithProgramVariables(var_ns, s);
		} catch (Exception e) {
			if (s.length()>10) {
				s=s.substring(0,10)+"...";
			}
			throw new antlr.SemanticException("Reading Java block "
				+ s + " failed. "
				+e.getMessage());	
		}
	}
	javaReader.setSVNamespace(var_ns);
	return javaReader.readBlockWithEmptyContext(s);
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
	if (javaInfo==null) {
           javaInfo=javaReader.createJavaInfo(new Services());
        }
	return javaInfo;
    }
}
top {Taclet r;}
:r=taclet;

taclet returns [Taclet r]{ 
	Sequent ifSeq = Sequent.EMPTY_SEQUENT;
	Object  find = null;
	r = null;
	TacletBuilder b = null;
    boolean sameUpdatePrefix = false;
}
:name:IDENT 
		LBRACE
		( "if" LPAREN ifSeq=seq RPAREN ) ?
		( "find" LPAREN find=termorseq RPAREN
		  ( "sameUpdateLevel" { sameUpdatePrefix = true; } ) ? ) ?
		{ 
			b = createTacletBuilderFor(find, sameUpdatePrefix);
			b.setName(new Name(name.getText()));
			b.setIfSequent(ifSeq);
		}
		( "varcond" LPAREN varexplist[b] RPAREN ) ?
		goalspecs[b]
		modifiers[b]
		RBRACE
		{ r = b.getTaclet(); }
    ;

modifiers[TacletBuilder b] :( heuristics[b] 
    | "noninteractive" { 
//      b.setNoninteractive(true);  
// "noninteractive" (as it is now) is confusing
// dropped this completely until a better solution (->Uwe)
      }
    | "recursive"      { b.setRecursive(true); }
    | DISPLAYNAME dname:STRING_LITERAL 
                       {b.setDisplayName(dname.getText());}
    | HELPTEXT htext:STRING_LITERAL
                       {b.setHelpText(htext.getText());}
    ) *
    ;

seq returns [Sequent s]{Semisequent ant,suc; s = null; }
:ant=semisequent SEQARROW suc=semisequent 
    { s = Sequent.createSequent(ant, suc); }
    ;

termorseq returns [Object o]{
	Term head = null;
	Sequent s = null;
	Semisequent ss = null;
	o = null; 
}
:head=term ( COMMA s=seq | SEQARROW ss=semisequent ) ?
		{
			if ( s == null ) {
				if ( ss == null ) {
					// Just a term
					o = head;
				} else {
					// A sequent with only head in the antecedent.
					Semisequent ant = Semisequent.EMPTY_SEMISEQUENT;
					ant = ant.insertFirst(
						new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
					o = Sequent.createSequent(ant,ss);
				}
			} else {
				// A sequent.  Prepend head to the antecedent.
				Semisequent newAnt = s.antecedent();
				newAnt = newAnt .insertFirst(
					new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
				o = Sequent.createSequent(newAnt,s.succedent());
			}
		}
    |
		SEQARROW ss=semisequent
		{
			o = Sequent.createSequent(Semisequent.EMPTY_SEMISEQUENT,ss);
		}
    ;

semisequent returns [Semisequent ss]{ 
	Term head = null;
	ss = Semisequent.EMPTY_SEMISEQUENT; 
}
:/* empty */ | 
		head=term ( COMMA ss=semisequent ) ? 
		{ ss = ss.insertFirst(new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent(); }
    ;

varexplist[TacletBuilder b] :varexp[b] ( COMMA varexp[b] ) * ;

varexp[TacletBuilder b] {
	ParsableVariable x;
	ParsableVariable y;
	Sort s = null;
}
:(( "typeof" LPAREN y=var_id  RPAREN  x=var_id "new")
			{ b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, false); }
		| ( "elemtypeof" LPAREN y=var_id  RPAREN  x=var_id "new")
			{ b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, true); }
		| (s=sort_id_check[true]  x=var_id "new") 
			{b.addVarsNew((SchemaVariable) x, s); }
		| (x=var_id "not" "free" "in" y=var_id)
			{ b.addVarsNotFreeIn((SchemaVariable)x,(SchemaVariable)y); }
		| (x=var_id "new" "depending" "on" y=var_id)
			{ b.addVarsNewDependingOn((SchemaVariable)x,(SchemaVariable)y); }
		)
    ;

goalspecs[TacletBuilder b] :"close" "goal"
    | goalspec[b] ( SEMI goalspec[b] )* ;

goalspec[TacletBuilder b] {
	Object rwObj = null;
	Sequent addSeq = Sequent.EMPTY_SEQUENT;
	ListOfTaclet addRList = SLListOfTaclet.EMPTY_LIST;
	SetOfSchemaVariable addpv = SetAsListOfSchemaVariable.EMPTY_SET;
        String name = null;
}
:(   ( "replacewith" (id:STRING_LITERAL {name = id.getText();})?
                      LPAREN rwObj=termorseq RPAREN
                         (addSeq=add)? 
                         (addRList=addrules)? 
			 (addpv=addprogvar)?
                    )
		| ( addSeq=add (addRList=addrules)? )
		| ( addRList=addrules )
		)
		{
			addGoalTemplate(b,name,rwObj,addSeq,addRList,addpv);
		}
    ;

add returns [Sequent s]{ s = null;}
:"add" LPAREN s=seq RPAREN;

addrules returns [ListOfTaclet lor]{ lor = null; }
:"addrules" LPAREN lor=tacletlist RPAREN;

addprogvar returns [SetOfSchemaVariable pvs]{pvs = null; }
:"addprogvars" LPAREN pvs=pvset RPAREN;

tacletlist returns [ListOfTaclet lor]{ 
	Taclet head = null;
	lor = SLListOfTaclet.EMPTY_LIST; 
}
:head=taclet   
		( /*empty*/ | COMMA lor=tacletlist ) { lor = lor.prepend(head); }
    ;

pvset returns [SetOfSchemaVariable pvs]{
        ParsableVariable pv = null;
	pvs = SetAsListOfSchemaVariable.EMPTY_SET;
}
:pv=var_id
		( /*empty*/ | COMMA pvs=pvset ) { pvs = pvs.add
							((SchemaVariable)pv); };

heuristics[TacletBuilder b] :"heuristics" LPAREN heuristic[b] ( COMMA heuristic[b] ) * RPAREN ;

heuristic[TacletBuilder b] :id:IDENT
		{   
			Heuristic h = (Heuristic) heur_ns.lookup(new Name(id.getText()));
			if (h == null) {
				throw new NotDeclException("heuristic", id,
							   getFilename());
			}
			b.addHeuristic(h);
		}
	;

bindingterm returns [Term a = null]:a = sv_or_local_bindingterm
	;

sv_or_local_bindingterm returns [Term a = null]{
	Operator op = null;
	QuantifiableVariable v = null;
	TermSymbol ts = null;
	Term a1;
	Sort s = null;
	boolean bound = false;
}
:(( 
            "all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)
		v0:IDENT
		(   
			COLON s=sort_id 
			{
				v=bindVar(v0,s);
				bound = true;
			}
		|
			/*empty*/
			{
				ts = (TermSymbol) var_ns.lookup(new Name(v0.getText()));	
				if (   ts == null
				    || ts instanceof LogicVariable ) {
					throw new SemanticException(
						"Quantified variables need a sort.");
				}
				if ( ! (ts instanceof SchemaVariable && 
					((SchemaVariable)ts).isVariableSV())) {
					throw new SemanticException(
						""+ts+" is not allowed in a quantifier.");
				}
				v = (SchemaVariable) ts;
				bound = false;
			}
		)
		DOT a1=term60 
		{ 
			a = tf.createQuantifierTerm((Quantifier)op, v, a1);
			if (bound) {unbindVars();}
		})
	| 
	
	a = update_or_substitution
	;

update_or_substitution returns [Term result = null]{
}
:( LBRACE attrterm ASSIGN) =>result = updateterm | 
    ( LBRACE IDENT (LBRACKET funcpredvarterm RBRACKET)* ASSIGN) => result = updateterm |
    result = substitutionterm 
  ;

updateterm returns [Term result = null]{ Term a0 = null;
  Term a1, a2;
  Vector subs = new Vector();  
  Vector index = new Vector();
  TermSymbol pv,pv1;
  ArraySort as;
  int n=0;
 }
:LBRACE	
   ( ( IDENT (LBRACKET IDENT RBRACKET)* DOT IDENT ) => a0 = attrterm 
      | v0:IDENT(LBRACKET v1:IDENT RBRACKET {pv1 = progVar(v1, (Sort) sort_ns.lookup(new Name("int")));
                                            index.add(tf.createFunctionTerm(pv1));n++;})*)
	ASSIGN a1=term60 {
                if(n>0){
                    as=ArraySortImpl.getArraySort(a1.sort());
                    for(int i=1;i<n;i++){
                        as=ArraySortImpl.getArraySort(as);
                    }
                    pv = progVar(v0, as);
                    a0 = tf.createFunctionTerm(pv);
                    Term[] t1 = (Term[]) index.toArray(new Term[index.size()]);
                    a0 = tf.createArrayTerm(Op.ARRAY,a0,t1);
                    n=0;
				}else if (a0 == null) {
			     	pv = progVar(v0, a1.sort()); 
	 			    a0 = tf.createFunctionTerm(pv); 
                }
		  	   subs.add(a0); 
	  		   subs.add(a1); 
		         }
     ( COMMA a0 = term60 ASSIGN a1=term60 {
  					    subs.add(a0); 
					    subs.add(a1);
					  } )* 
  RBRACE a2 = term60 {subs.add(a2);}
   { 
	result = tf.createUpdateTerm((Term[])subs.toArray(new Term[subs.size()]));
	subs = new Vector();
   }
;

substitutionterm returns [Term result = null]{ Term a1, a2;
  Sort s = null;
  TermSymbol ts;
  QuantifiableVariable v = null;
  boolean bound = false;
  boolean waryAll = false;
  boolean waryEx = false;
}
:LBRACE v1:IDENT 
    (COLON s=sort_id {
		       bound = true;
		     } | 
     { func_ns.lookup(new Name(v1.getText())) == null }?				
   	{
	  ts = (TermSymbol) var_ns.lookup(new Name(v1.getText()));
	  if (   ts == null
	      || ts instanceof LogicVariable ) {
	    throw new SemanticException("Substituted variables need a sort.");
	  }
	  if ( ! (ts instanceof SchemaVariable &&
		((SchemaVariable)ts).isVariableSV()) ) {
			throw new SemanticException(
			"	"+ts+" is not allowed as bound variable "
				+"in a substitution term.");
	  }
	  v = (SchemaVariable) ts;
	  bound = false;
       })    
    a1=term
	( "waryAll" { waryAll = true; } |
	  "waryEx"  { waryEx  = true; } )?
	RBRACE
     { 
	if (bound) {v = bindVar(v1,s);} 
     } 
    a2=term60
     { 
		    if ( waryAll )
		        result = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
		                                             v, a1, a2 );
		    else if ( waryEx )
		        result = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
		                                             v, a1, a2 );
		    else
		        result = tf.createSubstitutionTerm ( Op.SUBST,
		                                             v, a1, a2 );
	if (bound) {unbindVars();}
     }
;

attrterm returns [Term a = null]{	Vector vec=new Vector();
	ParsableVariable v=null;
	Term t=null;
	ParsableVariable v0=null;
    Sort so=null;

}
:(t=funcpredvarterm {	so=t.sort();
        } (DOT^ v=attrid[so] {vec.add(v); if(v!=null) so=v.sort(); })+) {
	if (!schematicJava) {
		for (int i=0; i<vec.size(); i++) {
			if(vec.get(i)!=null){
                t=tf.createAttributeTerm((ProgramVariable)vec.get(i), t);
            }
		}
	} else {
           t=tf.createAttributeTerm((SchemaVariable)v,t);         
	}
	a=t;
    };

attrid[Sort s] returns [ParsableVariable v = null]{

}
:attr:IDENT 
{ if (!schematicJava) {
		v = getJavaInfo().getAttribute(attr.getText(), (ObjectSort)s);
	} else {
		v = (SchemaVariable)var_ns.lookup(new Name(attr.getText()));
	}
  if ( v == null && !("length").equals(attr.getText()) ) {
     	throw new NotDeclException
		("Attribute ", attr.getText(),getFilename(), 
		  getLine(), getColumn());

  }

};

condterm returns [Term a = null]{ Term cond1_1 = null;
  Term cond1_2 = null;
  Term cond2_1 = null;
  Term cond2_2 = null;
  Term c_then = null;
  Term c_else = null;
  ParsableVariable v=null;
}
:(LPAREN term80 COMMA)=>(
	    LPAREN cond1_1=term80 COMMA cond1_2=term80 COLON c_else=term60 GUARD 
	cond2_1=term80 COMMA cond2_2=term80 COLON c_then=term60 RPAREN 
   {
	a = tf.createConjCondTerm(new Term[]{cond1_1, cond1_2, c_else, 
	  cond2_1, cond2_2, c_then});
   }
   )
   | LPAREN cond1_1=term80 GUARD cond2_1=term80 RPAREN 
    (

	DOT v = var_id CONDUP c_then = term60 {
		  a = tf.createIfElseTerm((SchemaVariable)v, cond1_1, 
					  cond2_1, c_then);
	        }	
	|
	c_then = term60 COLON c_else = term60 {
		  a = tf.createIfElseTerm(cond1_1, cond2_1, c_then, c_else);
		}
     ) 
;

// inherited from grammar TermParser
formula returns [Term a = null]:a = term 
		{
			if ( a.sort() != Sort.FORMULA ) {
				throw new antlr.SemanticException(
					"Just Parsed a Term where a Formula was expected");
			}
		}
	;

// inherited from grammar TermParser
term returns [Term a = null]:a=term20 
	;

// inherited from grammar TermParser
term20 returns [Term a = null]{Term a1;}
:a=term30 
		(EQV a1=term20 
			{ a = tf.createJunctorTerm(Op.EQV,new Term[]{a,a1});} ) ?;

// inherited from grammar TermParser
term30 returns [Term a = null]{Term a1;}
:a=term40 
		(IMP a1=term30 
			{ a = tf.createJunctorTerm(Op.IMP,new Term[]{a,a1});} ) ?;

// inherited from grammar TermParser
term40 returns [Term a = null]{Term a1;}
:a=term50 
		(OR a1=term40 
			{ a = tf.createJunctorTerm(Op.OR, new Term[]{a,a1});} ) ?;

// inherited from grammar TermParser
term50 returns [Term a = null]{Term a1;}
:a=term60 
		(AND a1=term50 
			{ a = tf.createJunctorTerm(Op.AND,new Term[]{a,a1});} ) ?;

// inherited from grammar TermParser
term60 returns [Term a = null]{
    Term a1,a2;
	Sort s;
}
:(LPAREN IDENT (COMMA IDENT COLON IDENT)? GUARD) => a=condterm
	|   a = term70
	|   a=bindingterm
	|   a=dl_term
	|   NOT a1=term60 { a = tf.createJunctorTerm(Op.NOT,new Term[]{a1}); }
	|   CTRL a1 = term60 { a = tf.createJunctorTerm(Op.CTRL,new Term[]{a1}); }

;

// inherited from grammar TermParser
term70 returns [Term a = null]{Term a1;}
:a=term80
		(EQUALS a1=term80
			{ a = tf.createEqualityTerm(a,a1);} ) ?;

// inherited from grammar TermParser
term80 returns [Term a = null]:(IDENT (LBRACKET funcpredvarterm RBRACKET)* DOT IDENT)=> a=attrterm
    |   AT a = abbreviation
    |   ARITH_OP {minus = !minus; a=term80();}
    |   a = funcpredvarterm
    |   LPAREN a = term RPAREN
    |   "true" { a = tf.createJunctorTerm(Op.TRUE); }
    |   "false" { a = tf.createJunctorTerm(Op.FALSE); }
    ;
exception 
catch [RecognitionException ex] {
        throw new RecognitionException("Not a valid logical term.\n"+ex.toString());
    }

// inherited from grammar TermParser
abbreviation returns [Term a=null]{ 
	String sc = null;
}
:(	sc = id
		{
			a =  scm.getTerm(sc);
		    if(a==null){
				throw new NotDeclException
				("abbreviation", sc, 
				getFilename(), getLine(), getColumn());
			}
			
                
		}
)
        ;

// inherited from grammar TermParser
local_bindingterm returns [Term a = null]{
	Operator op = null;
	LogicVariable v = null;
	Term a1;
	Sort s;
}
:(
			"all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)

		v0:IDENT COLON s=sort_id DOT {v = bindVar(v0,s);} a1=term60
		{
			a = tf.createQuantifierTerm((Quantifier)op, v,a1);
			unbindVars();
		}
	| a = update_or_substitution
	;

// inherited from grammar TermParser
dl_term returns [Term a = null]:a = normal_dl_term
	;

// inherited from grammar TermParser
normal_dl_term returns [Term a = null]{	JavaBlock jb=null;
	Term a1;
}
:d1:DIAMOND 
		{
			jb=getJavaBlock(d1);
			HashSet progVars=progVars(jb);
			bindProgVars(progVars);
		}
		        a1=term60	
		{
			a = tf.createDiamondTerm(jb, a1);
			unbindProgVars();
		}
	|
		d2:BOX 
		{
			jb=getJavaBlock(d2);
			HashSet progVars=progVars(jb);
			bindProgVars(progVars);
		}
		        a1=term60	
		{
			a = tf.createBoxTerm(jb, a1);
			unbindProgVars();
		}
    ;

// inherited from grammar TermParser
funcpredvarterm returns [Term a = null]{
	Operator vf;
	int n = 0;
	Vector al = null;
	Term t;
	Sort as=null;
}
:({
		String s=LT(0)+" ";
		s=LT(1)+" "; 
		s=LT(2)+" ";
		s=LT(3)+" ";        
		if(LT(1)!=null && isNumber(LT(1).getText())){
			if(LT(0)==null || LT(2)==null){
			    a=toZNotation(LT(1).getText(), minus);
			    minus=false;  		
			    match(IDENT);
			    return a;
			}else if(!LT(0).getText().equals("(")|| !LT(2).getText().equals("(")){ 
			    a=toZNotation(LT(1).getText(), minus);
			    minus=false;  		
			    match(IDENT);
			    return a;
			}
		}  
		}
		|		
		 CHAR_LITERAL {
				String s=LT(0).getText();
				int intVal=0;
				if (s.length()==3) {
					intVal=(int)LT(0).getText().charAt(1);
				} else {
					try {
					intVal = Integer.parseInt(s.substring(3,s.length()-1),16);
					} catch (NumberFormatException ex) {
					 throw new KeYSemanticException
					   ("Expected '"+s+"' to be a character", LT(0), getFilename());
					} 		
				}
				t=toZNotation(""+intVal,false);
		     		vf = (Function) func_ns.lookup(new Name("C"));
				return a=tf.createFunctionTerm((Function)vf, t.sub(0));
			      }
 		) 
		vf = varfunc_id
		(
			{ vf.arity() > 0 }?
				LPAREN 
				t=term 
				{
					al = new Vector();
					n=1;
					al.add(t);
				}
				( COMMA 
					t=term 
					{
						n++;
						al.add(t);
					}
				)* RPAREN	

                )?
                (
                 (LBRACKET)=>  
                   {vf instanceof ParsableVariable}? 
		   {a = termForParsedVariable((ParsableVariable)vf);}
		   {as = a.sort();} 
            {as instanceof ArraySort || as == ProgramSVSort.VARIABLE}?
                   (LBRACKET
                       t=term60
                       {
                         if (as!=null && as!=ProgramSVSort.VARIABLE){
                           if (!(as instanceof ArraySort)){
                                    throw new NotDeclException
                                    ("array "+vf.name()+" declared with a different"+
                                    " dimension than required here", vf.name().toString(), 
                                    getFilename(), getLine(), getColumn());
                           } else {
                                    as = ((ArraySort) as).elementSort();
                           } 
                         }
                         a=tf.createArrayTerm(Op.ARRAY,a,t);
                       }
                     RBRACKET)+                 
                |                
		 {  if (vf instanceof ParsableVariable) {
		      a = termForParsedVariable((ParsableVariable)vf);
		    } else if (vf instanceof Function) {
			if ( al == null ) {	
				a = tf.createFunctionTerm((Function)vf);
			} else {
				a = tf.createFunctionTerm((Function)vf,
					   (Term[])al.toArray(AN_ARRAY_OF_TERMS));
				}

		    } else if (vf instanceof MetaOperator) {

			if ( al == null ) {	
				a = tf.createMetaTerm((MetaOperator)vf, new Term[0]);
			} else {	
				a = tf.createMetaTerm((MetaOperator)vf,
					   (Term[])al.toArray(AN_ARRAY_OF_TERMS));
			}
		    }
		    if (minus) {
	              Term[] t1 = new Term[1];
        	      t1[0] = a;
               	      vf = (Function) func_ns.lookup(new Name("~m"));
                      a = tf.createFunctionTerm((Function)vf,t1);
                      minus = false;
            	    }	
	         }
		)
		;

// inherited from grammar TermParser
varfunc_id returns [Operator v = null]{ 
	String varfunc_name = null;
}
:(	v = expr_op	
		| varfunc_name = id
		{
			v = (TermSymbol) var_ns.lookup(new Name(varfunc_name));
			if (v == null) {
			   v = (Function) func_ns.lookup(new Name(varfunc_name));
			   if(v==null) {
                           	v = (ProgramVariable)
                                	getJavaInfo().programVariables().lookup(
                                        	new Name(varfunc_name));
                                if(v==null) {
					throw new NotDeclException
					("variable or function", varfunc_name, 
					getFilename(), getLine(), getColumn());
                                }
			   }
			}
                
		}
		| META_INT_UNARY_MINUS 
		{ v = Op.META_INT_UNARY_MINUS ; }
		| META_LONG_UNARY_MINUS 
		{ v = Op.META_LONG_UNARY_MINUS ; }
		| META_ADD 
		{ v = Op.META_ADD; }
		| META_SUB
		{ v = Op.META_SUB; }
		| META_MUL
		{ v = Op.META_MUL; }
		| META_DIV 
		{ v = Op.META_DIV; }
		| META_MOD
		{ v = Op.META_MOD; }
		| META_LESS
		{ v = Op.META_LESS; }
		| META_GREATER
		{ v = Op.META_GREATER; }
		| META_LEQ
		{ v = Op.META_LEQ; }
		| META_GEQ
		{ v = Op.META_GEQ; }
		| META_EQ
		{ v = Op.META_EQ; }
		| META_INT_ADD 
		{ v = Op.META_INT_ADD; }
		| META_INT_SUB 
		{ v = Op.META_INT_SUB; }
		| META_INT_MUL 
		{ v = Op.META_INT_MUL; }
		| META_INT_DIV 
		{ v = Op.META_INT_DIV; }
		| META_INT_MOD 
		{ v = Op.META_INT_MOD; }
		| META_INT_AND 
		{ v = Op.META_INT_AND; }
		| META_INT_OR
		{ v = Op.META_INT_OR; }
		| META_INT_XOR
		{ v = Op.META_INT_XOR; }
		| META_INT_COMPLEMENT
		{ v = Op.META_INT_COMPLEMENT; }
		| META_INT_SHIFTRIGHT
		{ v = Op.META_INT_SHIFTRIGHT; }
		| META_INT_SHIFTLEFT
		{ v = Op.META_INT_SHIFTLEFT; }
		| META_INT_UNSIGNEDSHIFTRIGHT
		{ v = Op.META_INT_UNSIGNEDSHIFTRIGHT; }
		| META_LONG_ADD 
		{ v = Op.META_LONG_ADD; }
		| META_LONG_SUB 
		{ v = Op.META_LONG_SUB; }
		| META_LONG_MUL 
		{ v = Op.META_LONG_MUL; }
		| META_LONG_DIV 
		{ v = Op.META_LONG_DIV; }
		| META_LONG_MOD 
		{ v = Op.META_LONG_MOD; }
		| META_LONG_AND 
		{ v = Op.META_LONG_AND; }
		| META_LONG_OR
		{ v = Op.META_LONG_OR; }
		| META_LONG_XOR
		{ v = Op.META_LONG_XOR; }
		| META_LONG_COMPLEMENT
		{ v = Op.META_LONG_COMPLEMENT; }
		| META_LONG_SHIFTRIGHT
		{ v = Op.META_LONG_SHIFTRIGHT; }
		| META_LONG_SHIFTLEFT
		{ v = Op.META_LONG_SHIFTLEFT; }
		| META_LONG_UNSIGNEDSHIFTRIGHT
		{ v = Op.META_LONG_UNSIGNEDSHIFTRIGHT; }
		| META_IN_LONG 
		{ v = Op.META_IN_LONG; }
		| META_IN_INT
		{ v = Op.META_IN_INT; }
		| META_IN_CHAR
		{ v = Op.META_IN_CHAR; }
		| META_IN_SHORT
		{ v = Op.META_IN_SHORT; }
		| META_IN_BYTE 
		{ v = Op.META_IN_BYTE; }
		| META_MODULO_BYTE
		{ v = Op.META_MODULO_BYTE; }
		| META_MODULO_SHORT
		{ v = Op.META_MODULO_SHORT; }
		| META_MODULO_INTEGER
		{ v = Op.META_MODULO_INTEGER; }
		| META_MODULO_LONG 
		{ v = Op.META_MODULO_LONG; }	
		| STATIC_ANALYSIS
		{ v = Op.STATIC_ANALYSIS; }
		| META_LENGTH
		{ v = Op.META_LENGTH; })
	;

// inherited from grammar TermParser
expr_op returns [Operator v = null]{
	ParsableVariable left_var0 = null;
	ParsableVariable right_var0 = null;
	ParsableVariable unary_var0 = null;
}
:IN_TYPE LPAREN ((left_var0=var_id ((id:ARITH_OP | id1:SLASH | id2:IDENT) right_var0 = var_id)?) |
			   ( "-" unary_var0 = var_id) )  RPAREN
	{  
           if (id==null) id=id1;	
	   if (unary_var0!=null) {
	      // can only be unary minus
	      SchemaVariable unary_var = (SchemaVariable) unary_var0;
	      Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
	      return new InTypeOperator
		(type, new de.uka.ilkd.key.java.expression.operator.Negative(unary_var));
	   }
	   else {
	      SchemaVariable left_var = (SchemaVariable) left_var0;
	      Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
	      if (right_var0 != null) {
	         SchemaVariable right_var = (SchemaVariable) right_var0;
    	         String op_str = (id == null ? id2.getText() : id.getText());
           	   if ("+".equals(op_str)) {
		     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Plus(left_var, right_var));
	         } else if ("-".equals(op_str)) {
		     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Minus(left_var, right_var));
	         } else if ("*".equals(op_str)) {
		     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Times(left_var, right_var));
	         } else if ("/".equals(op_str)) {
	    	     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Divide(left_var, right_var));
		} else if ("%".equals(op_str)) {
	    	     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Modulo(left_var, right_var));
  	         } 
	      }
	      else {
	         return new InTypeOperator(type, left_var);
	      }
 	   }
	}
	| IS_OBJECT  LPAREN (left_var0=var_id )  RPAREN
	{
 	   SchemaVariable var = (SchemaVariable) left_var0;
	   return new IsObject(de.uka.ilkd.key.logic.sort.Sort.FORMULA, var);
	}
	| IS_CLONABLE  LPAREN (left_var0=var_id )  RPAREN
	{
 	   SchemaVariable var = (SchemaVariable) left_var0;
	   return new IsClonable(de.uka.ilkd.key.logic.sort.Sort.FORMULA, var);
	}
;

// inherited from grammar TermParser
var_id returns [ParsableVariable v = null]:id:IDENT 
		{   
			v = (ParsableVariable) var_ns.lookup(new Name(id.getText()));
			if (v == null) {
				throw new NotDeclException("variable", id, 
							   getFilename());
			}
		}
	;

// inherited from grammar DeclParser
decls :(
			sort_decls
		|
			schema_var_decls
		|
                	prog_var_decls
                |
			pred_decls
		|
			func_decls
		|
			heur_decls
		) *
	;

// inherited from grammar DeclParser
sort_decls {
    boolean isObjectSort=false;
    boolean isGenericSort=false;
    Sort[] sortExt=new Sort [0];
    Sort[] sortOneOf=new Sort [0];
    ListOfName sort_ids;
}
:"sorts" 
	LBRACE ( 
           { 
	       isObjectSort=false; 
	       isGenericSort=false; 
	       sortExt = new Sort[0];
	       sortOneOf = new Sort[0];
	   }
	  ( "object" {isObjectSort=true;} sort_ids = sort_identifiers
	  | "generic" {isGenericSort=true;} sort_ids = sort_identifiers
	  ( "oneof" sortOneOf = oneof_sorts )? 
	  ("extends" sortExt = extends_sorts )?
	  | sort_ids = sort_identifiers
      )
	  SEMI
	{
	    Name           sort_name;
	    IteratorOfName it         = sort_ids.iterator ();
	    while ( it.hasNext () ) {
	    sort_name = it.next ();

	    if (sort_ns.lookup(sort_name) == null) {
		Sort s;
		if (isObjectSort) {
		    if ("Object".equals(sort_name)) {
			if (sort_ns.lookup(new Name("Object"))
			    ==null) {

			    s = new ClassInstanceSortImpl(sort_name);
			}
			s=(Sort)sort_ns.lookup(new Name("Object"));
		    } else {
			s = new ClassInstanceSortImpl(sort_name, 
						      (Sort)sort_ns.lookup
						      (new Name("Object")));
		    }	
		} else if (isGenericSort) {
		    int        i;
		    SetOfSort  ext   = GenericSort.EMPTY_SORT_SET;
		    SetOfSort  oneOf = GenericSort.EMPTY_SORT_SET;

		    for ( i = 0; i != sortExt.length; ++i )
		        ext = ext.add ( sortExt[i] );

		    for ( i = 0; i != sortOneOf.length; ++i )
		        oneOf = oneOf.add ( sortOneOf[i] );

		    try {
		        s = new GenericSort ( sort_name,
		                              ext,
		                              oneOf );
		    } catch ( GenericSupersortException e ) {
		        throw new GenericSortException ( "sort",
		                   "Illegal sort given", e.getIllegalSort (),
		                   getFilename (), getLine (), getColumn () );
		    }
		} else {
		    s = new PrimitiveSort(sort_name);
		}
		sort_ns.add(s);
		if (s instanceof NonCollectionSort) {
		    NonCollectionSort ns
			= (NonCollectionSort)s;
		    sort_ns.add(ns.getSetSort());
		    sort_ns.add(ns.getSequenceSort());
		    sort_ns.add(ns.getBagSort());
		}
		if ( s instanceof SortDefiningSymbols )
		    ((SortDefiningSymbols)s).createSymbols ( func_ns, sort_ns );
	    }

	    }
	}
			
	)* 
	RBRACE
	;

// inherited from grammar DeclParser
sort_identifiers returns [ListOfName ids = SLListOfName.EMPTY_LIST]:id:IDENT { ids = ids.append ( new Name(id.getText()) ); }
        (
            COMMA
            id2:IDENT { ids = ids.append ( new Name(id2.getText()) ); }
        ) *
    ;

// inherited from grammar DeclParser
extends_sorts returns [Sort[] extendsSorts = null]{
	Vector args = new Vector();
	Sort s;
}
:s = any_sort_id_check[true] { args.add(s); }
        (
            COMMA s = any_sort_id_check[true] {args.add(s);}
        ) *
		{
			extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;

// inherited from grammar DeclParser
oneof_sorts returns [Sort[] oneOfSorts = null]{
	Vector args = new Vector();
	Sort s;
}
:LBRACE
        s = sort_id_check[true] { args.add(s); }
        (
            COMMA s = sort_id_check[true] {args.add(s);}
        ) *
        RBRACE
		{
			oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;

// inherited from grammar DeclParser
schema_var_decls {
	Sort s = null;
	boolean makeVariableSV  = false;
	boolean makeDependingSV = false; // skolem functions
	boolean rigid           = false;
	boolean listModifier    = false;
}
:"schema" "variables"
		LBRACE
		(
			{makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;}
			( ( "program" ("svlist" {listModifier=true;})?
					( "statement"
					{
						s=ProgramSVSort.STATEMENT;
					}
					| "expression"
					{
						s=ProgramSVSort.EXPRESSION;
					}
					| "multiplevardecl"
					{
						s=ProgramSVSort.MULTIPLEVARDECL;
					}
					| "simple" "expression"
					{
						s=ProgramSVSort.SIMPLEEXPRESSION;
					}
					| "simple" "javaboolean" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABOOLEXPRESSION;
					}
					| "simple" "javabyte" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABYTEEXPRESSION;
					}
					| "simple" "javachar" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVACHAREXPRESSION;
					}
					| "simple" "javashort" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVASHORTEXPRESSION;
					}
					| "simple" "javaint" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVAINTEXPRESSION;
					}
					| "simple" "javalong" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVALONGEXPRESSION;
					}
					| "simple" "javabyteshortint" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABYTESHORTINTEXPRESSION;
					}
					| "simple" "anyjavatype" "expression"
					{
						s=ProgramSVSort.SIMPLEANYJAVATYPEEXPRESSION;
					}
					| "classtype" "expression"
					{
						s=ProgramSVSort.CLASSTYPEEXPRESSION;
					}
					| "interfacetype" "expression"
					{
						s=ProgramSVSort.INTERFACETYPEEXPRESSION;
					}
					| "nonsimple" "expression"	
					{
						s=ProgramSVSort.NONSIMPLEEXPRESSION;
					}
					| "javaboolean" "expression"	
					{
						s=ProgramSVSort.BOOLEANEXPRESSION;
					}
					| "nonsimple" "javaboolean" "expression"	
					{
						s=ProgramSVSort.NONSIMPLEBOOLEANEXPRESSION;
					} 
					| "lefthandside"
					{
						s=ProgramSVSort.LEFTHANDSIDE;
					}
					| "loopinit"
					{
						s=ProgramSVSort.LOOPINIT;
					}
					| "guard"
					{
						s=ProgramSVSort.GUARD;
					}
					| "forupdates"
					{
						s=ProgramSVSort.FORUPDATES;
					}
					| "statementexpression"
					{
						s=ProgramSVSort.STATEMENTEXPRESSION;
					}
					| "label"
					{
						s=ProgramSVSort.LABEL;
					}
					| "type"
					{
						s=ProgramSVSort.TYPE;
					}
					| "variable"	
					{
						s=ProgramSVSort.VARIABLE;
					}
					| "variableinitializer"	
					{
						s=ProgramSVSort.VARIABLEINIT;
					}
					| "implicit" "next"	
					{
						s=ProgramSVSort.IMPLICITNEXT;
					}				
					| "implicit" "reference" "field"	
					{
						s=ProgramSVSort.IMPLICITREFERENCE;
					}				
					| "new"	
					{
						s=ProgramSVSort.NEW;
					}
					| "new"	"array"
					{
						s=ProgramSVSort.NEWARRAY;
					}
					| "array" "initializer"
					{
						s=ProgramSVSort.ARRAYINITIALIZER;
					}
					| "array" "length"
					{
						s=ProgramSVSort.ARRAYLENGTH;
					}
                    			| "switch"
                    			{
                        		 	s=ProgramSVSort.SWITCH;
                    			}
					|		
					( "booleanvariable") {
						s=ProgramSVSort.BOOLEANVARIABLE;
					}
					|
					( "catch") {
						s=ProgramSVSort.CATCH;
					}
					|		
					( "methodreference") {
						s=ProgramSVSort.METHODREFERENCE;
					}
					|
					  "special" "constructorreference" {
						s=ProgramSVSort.SPECIALCONSTRUCTORREFERENCE;				
					}	
					|		
					( "classtype") {
						s=ProgramSVSort.CLASSTYPE;
					}
					|		
					( "interfacetype") {
						s=ProgramSVSort.INTERFACETYPE;
					}
					|		
					( "arraytype") {
						s=ProgramSVSort.ARRAYTYPE;
					}
					|		
					( "array") {
						s=ProgramSVSort.ARRAY;
					}
					|
					( "methodbody") {
						s=ProgramSVSort.METHODBODY;
					}
					|
					( "branch") {
						s=ProgramSVSort.BRANCH;
					}
					)
				)
				|
				"formula"
                		( "rigid" { rigid = true; } ) ?
				{ 
					s = Sort.FORMULA;
				}			
				|
				s=any_sort_id_check[true]
				( "variables" { makeVariableSV = true; }
                		| "depending" { makeDependingSV = true; }
                		| "rigid" { rigid = true; } ) ?
				
			)
			one_schema_var_decl[s,makeVariableSV,makeDependingSV,rigid,listModifier]
			( COMMA one_schema_var_decl[s,makeVariableSV,makeDependingSV,rigid,listModifier] ) *
			SEMI
		) *
		RBRACE
	;

// inherited from grammar DeclParser
one_schema_var_decl[Sort s, 
	            boolean makeVariableSV, 
	            boolean makeDependingSV,
	            boolean rigid, 
	            boolean listModifier] :id:IDENT
		{
		    if (!skip_schemavariables) {
			SchemaVariable v;
			if ( s == Sort.FORMULA ) {
			    v = SchemaVariable.createFormulaSV
				(new Name(id.getText()), listModifier, rigid);
			} else if ( s instanceof ProgramSVSort ) {
			    v = SchemaVariable.createProgramSV
			    (new ProgramElementName(id.getText()),(ProgramSVSort) s, listModifier);
			} else {
			    if ( makeVariableSV ) {
				v=SchemaVariable.createVariableSV
				(new Name(id.getText()), s, listModifier);
			    } else if ( makeDependingSV ) {
				v = SchemaVariable.createDependingSV
				(new Name(id.getText()), s, listModifier);
			    } else { v = SchemaVariable.createTermSV
				     (new Name(id.getText()), s, listModifier, rigid);
			    }
			}
			var_ns.add(v);
		    }
		}
	;

// inherited from grammar DeclParser
pred_decls {
	Sort[] argSorts;
	String pred_name;
}
:"predicates"
		LBRACE
		(
			pred_name = id
			argSorts=arg_sorts[!skip_predicates]
			{
			    if (!skip_predicates) {
				Name predicate = new Name(pred_name);
				if (func_ns.lookup(predicate) != null) {
				    // throw new AmbigiousDeclException
				    //("predicate "+pred_name+" ", 
				    //getLine(), getColumn());
				} else {
				    Function p = new Function(predicate,
							      Sort.FORMULA, 
							      argSorts);
				    func_ns.add(p);			
				}
			    } 
			}
			SEMI
		) *
		RBRACE
	;

// inherited from grammar DeclParser
func_decls {
	Sort[] argSorts;
	Sort retSort;
        String func_name;
}
:"functions"
		LBRACE
		(
			retSort = sort_id_check[!skip_functions]
			func_name = id
			argSorts = arg_sorts[!skip_functions]
			{
			    if (!skip_functions) {

				Name fct_name = new Name(func_name);
				if (func_ns.lookup(fct_name) != null) {
				    //throw new AmbigiousDeclException
				    //("function "+func_name, 
				    //getLine(), getColumn());
				} else {
				    Function f=new Function(fct_name,
							    retSort, 
							    argSorts);
				    func_ns.add(f);
				}
			    } 
			}
			SEMI
		) *
		RBRACE
	;

// inherited from grammar DeclParser
arg_sorts[boolean checkSort] returns [Sort[] argSorts = null]{
	Vector args = new Vector();
	Sort s;
}
:(
			LPAREN
			s = sort_id_check[checkSort] { args.add(s); }
			(
				COMMA s = sort_id_check[checkSort] {args.add(s);}
			) *
			RPAREN
		) ?
		{
			argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;

// inherited from grammar DeclParser
prog_var_decls {
	Sort s;
        String var_name;
}
:"program" "variables"
		LBRACE
		(
			s = sort_id_check[true]
			var_name = id
			{
				ProgramVariable pv = new ProgramVariable(
                                   new ProgramElementName(var_name), 
                                   javaInfo.getKeYJavaType(s));
				javaInfo.programVariables().add(pv);
			}
			SEMI
		) *
		RBRACE
	;

// inherited from grammar DeclParser
heur_decls :"heuristics"
		LBRACE
		(
			id:IDENT SEMI
			{ 
			    if (!skip_heuristics) {
				Heuristic h = new Heuristic(new Name(id.getText()));
				heur_ns.add(h);
			    }
			}
		) *
		RBRACE
	;

// inherited from grammar DeclParser
sort_id returns [Sort s = null]:s = sort_id_check[true]
	;

// inherited from grammar DeclParser
sort_id_check[boolean checkSort] returns [Sort s = null]{
    Sort t;
}
:t = sort_id_check_help[checkSort]
        s = array_decls[t]
	;

// inherited from grammar DeclParser
any_sort_id_check[boolean checkSort] returns [Sort s = null]{
    Sort t;
}
:t = any_sort_id_check_help[checkSort]
        s = array_decls[t]
	;

// inherited from grammar DeclParser
any_sort_id_check_help[boolean checkSort] returns [Sort s = null]{
    String name;
}
:id:IDENT
                { 	String idtext=(useInstead.get(id.getText()));
			s= (Sort) sort_ns.lookup(new Name(idtext));
			if (checkSort && s == null) {
			    throw 
				new NotDeclException("sort", id, 
						     getFilename());
			}
		}
	;

// inherited from grammar DeclParser
sort_id_check_help[boolean checkSort] returns [Sort s = null]:s = any_sort_id_check_help[checkSort]
        {
            // don't allow generic sorts or collection sorts of
            // generic sorts at this point
            Sort t = s;
            while ( t instanceof CollectionSort )
                t = ((CollectionSort)t).elementSort ();
            if ( t instanceof GenericSort )
                throw new GenericSortException ( "sort",
                        "Non-generic sort expected", t,
                        getFilename (), getLine (), getColumn () );
        }
    ;

// inherited from grammar DeclParser
array_decls[Sort p] returns [Sort s = null]{
    int  n=0;
}
:(LBRACKET RBRACKET {n++;})*


	        { s = p;
              while(n>0){
                   s=ArraySortImpl.getArraySort(s);
                   n--;
              }
		 } 
	;

// inherited from grammar DeclParser
id_declaration returns [ IdDeclaration idd = null ]{
    Sort s = null;
}
:id:IDENT
        ( COLON s = sort_id_check[true] ) ?
        {
            idd = new IdDeclaration ( id.getText (), s );
        }
    ;

// inherited from grammar DeclParser
id returns [String name = null]:(IDENT COLON COLON IDENT) => 
                	classID:IDENT COLON COLON nameID:IDENT 
			{ 
			 name = classID.getText() + "::" + nameID.getText();
			}
                |
		pureID:IDENT { name = pureID.getText(); } 
	;


